#include <boost/graph/subgraph.hpp>
#include <boost/graph/copy.hpp>
#include <boost/graph/depth_first_search.hpp>

#include <GraphAlgorithms.h>
#include <Generic.h>
#include <Message.h>
#include <GraphTreePartition.h>
#include <Potential.h>
#include <Node.h>
#include <RandomVariable.h>
#include <OutputGraph.h>
#include <Experiment.h>

using namespace boost;
using namespace std;

// Explicit Template Instantiations //

typedef property< edge_index_t, unsigned int, DiscretePotential * > DiscretePairwiseEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, MessageNode <DiscreteRandomVariable, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> *, DiscretePairwiseEdgeProperty > DiscretePairwiseGraph;
typedef subgraph <DiscretePairwiseGraph> SubDiscretePairwiseGraph;

typedef property< edge_index_t, unsigned int > EdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, MessageNode <DiscreteRandomVariable, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> *, EdgeProperty > DiscreteFactorGraph;
typedef subgraph <DiscreteFactorGraph> SubDiscreteFactorGraph;

//End of typedefs

template void belief_propagation < DiscretePairwiseGraph  > (DiscretePairwiseGraph &);
template void belief_propagation < DiscreteFactorGraph  > (DiscreteFactorGraph &);

template void loopy_belief_propagation < DiscretePairwiseGraph  > (DiscretePairwiseGraph &, const unsigned int, const bool = false);
template void loopy_belief_propagation < DiscreteFactorGraph  > (DiscreteFactorGraph &, const unsigned int, const bool = false);

template void gibbs_sampler < DiscretePairwiseGraph > (DiscretePairwiseGraph &, const unsigned int, const unsigned int, const bool);
template void gibbs_sampler < DiscreteFactorGraph  > (DiscreteFactorGraph &, const unsigned int, const unsigned int, const bool);

template void tree_gibbs_sampler < DiscretePairwiseGraph > (DiscretePairwiseGraph &, const unsigned int, const unsigned int = 0, const bool = true, const bool = false);
template void tree_gibbs_sampler < DiscreteFactorGraph > (DiscreteFactorGraph &, const unsigned int, const unsigned int = 0, const bool = true, const bool = false);

// End Explicit Template Instantiations //


//******** Helper functions (need to be specialized) ************//

// Generic version

// Version for the DiscretePairwiseGraph (and SubGraph, needed for the Tree Partition...)

template <>
inline void setup_message_computation(DiscretePairwiseGraph & g, const graph_traits<DiscretePairwiseGraph>::vertex_descriptor origin, const graph_traits<DiscretePairwiseGraph>::vertex_descriptor destination)
{
	static_cast <MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscretePairwiseGraph>::vertex_descriptor> *> (g[origin])->message_potential = g[edge (origin, destination, g).first ];	
}

template <>
inline void setup_message_computation(SubDiscretePairwiseGraph & g, const graph_traits<SubDiscretePairwiseGraph>::vertex_descriptor origin, const graph_traits<SubDiscretePairwiseGraph>::vertex_descriptor destination)
{
	static_cast <MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<SubDiscretePairwiseGraph>::vertex_descriptor> *> (g[origin])->message_potential = g[edge (origin, destination, g).first ];	
}

template <>
void setup_gibbs_proposal(DiscretePairwiseGraph & g, const graph_traits<DiscretePairwiseGraph>::vertex_descriptor node)
{
	graph_traits<DiscretePairwiseGraph>::out_edge_iterator current_edge, end_edge;
	
	MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscretePairwiseGraph>::vertex_descriptor> * mrf_node = static_cast <MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscretePairwiseGraph>::vertex_descriptor> *> (g[node]);
	
	mrf_node->sample_proposal = new ChainedSingleDiscretePotential(* mrf_node->variable_ptr);
	
	mrf_node->sample_proposal->add_potential( mrf_node->potential_ptr);
	
	for ( tie (current_edge, end_edge) = out_edges( node, g); current_edge != end_edge; ++current_edge)
	{
		mrf_node->sample_proposal->add_potential( g[*current_edge]);
	}
	
	mrf_node->variable_ptr->remove_prior_samples();
}

template <>
void setup_tree_gibbs_sampler(SubDiscretePairwiseGraph & g, vector < vector <double> > & sum_marginals, vector < Potential * > & original_potentials)
{
	property_map<SubDiscretePairwiseGraph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	
	graph_traits<SubDiscretePairwiseGraph>::vertex_iterator u, u_end;
	graph_traits<SubDiscretePairwiseGraph>::adjacency_iterator v, v_end;
	subgraph<DiscretePairwiseGraph>::children_iterator current_tree, tree_end;
	graph_traits<DiscretePairwiseGraph>::vertex_descriptor w;
	
	for ( tie(current_tree, tree_end) = g.children(); current_tree != tree_end; ++current_tree)
	{
		for (tie(u,u_end) = vertices (*current_tree); u!= u_end; ++u)
		{
			w = current_tree->local_to_global(*u);
			
			MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * node_casted = static_cast < MRFMessageNode<DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> (g[w]);
			
			sum_marginals[w] = vector < double> (node_casted->variable_ptr->get_number_values(),0.0);
			
			original_potentials[ get(vertex_index_map, w)] = static_cast < DiscretePotential *> (g[w]->get_potential());
			
			ChainedSingleDiscretePotential * p = new ChainedSingleDiscretePotential(* node_casted->variable_ptr);
			
			p->add_potential (static_cast < DiscretePotential *> (original_potentials[get(vertex_index_map, w)]));
			
			node_casted->potential_ptr = p;			
			
			for (tie(v, v_end)= adjacent_vertices ( w, g ); v != v_end; ++v)
			{
				if (!current_tree->find_vertex(*v).second)
				{
					static_cast < ChainedSingleDiscretePotential *> (g[w]->get_potential())->add_potential(g [edge(w, *v, g).first]);
				}
			}
			
		}
		
	}
	
}


template <>
void cleanup_tree_gibbs_sampler(DiscretePairwiseGraph & g, vector < Potential * > & original_potentials)
{
	property_map<DiscretePairwiseGraph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	graph_traits<DiscretePairwiseGraph>::vertex_iterator u, u_end;
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{		
		MRFMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * node_casted = static_cast < MRFMessageNode<DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> (g[*u]);
		
		delete node_casted->potential_ptr;
		node_casted->potential_ptr = static_cast <DiscretePotential *> (original_potentials[ get(vertex_index_map, *u)] );
	}	
}

template <>
void remove_temp_observations(SubDiscretePairwiseGraph &)
{
}

template <>
void set_temp_observations(SubDiscretePairwiseGraph &)
{
}

template <>
void initialize_graph_variables_random(DiscretePairwiseGraph & g)
{
	graph_traits<DiscretePairwiseGraph>::vertex_iterator u, u_end;
	graph_traits<DiscretePairwiseGraph>::out_edge_iterator e, e_end;
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		g[*u]->get_random_variable()->set_value_at_random();
		
		for (tie(e,e_end) = out_edges (*u, g); e!= e_end; ++e)
		{
			g[*e]->set_variable_value(* g[*u]->get_random_variable());	
		}
	}
	
}


// Version for the DiscreteFactorGraph (nothing needs to be done here)

template <>
inline void setup_message_computation(DiscreteFactorGraph &, const graph_traits<DiscreteFactorGraph>::vertex_descriptor, const graph_traits<DiscreteFactorGraph>::vertex_descriptor)
{
}

template <>
inline void setup_message_computation(SubDiscreteFactorGraph &, const graph_traits<SubDiscreteFactorGraph>::vertex_descriptor, const graph_traits<SubDiscreteFactorGraph>::vertex_descriptor)
{
}

template <>
void setup_gibbs_proposal(DiscreteFactorGraph & g, const graph_traits<DiscreteFactorGraph>::vertex_descriptor node)
{
	graph_traits<DiscreteFactorGraph>::adjacency_iterator current_neighbour, end_neighbour;
	
	VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[node]);
	
	if ( vmn != NULL)
	{
		vmn->sample_proposal = new ChainedSingleDiscretePotential(* vmn->variable_ptr);
		
		vmn->sample_proposal->add_potential( vmn->potential_ptr);
		
		for ( tie (current_neighbour, end_neighbour) = adjacent_vertices( node, g); current_neighbour != end_neighbour; ++current_neighbour)
		{
			vmn->sample_proposal->add_potential( static_cast < DiscretePotential * > (g[*current_neighbour]->get_potential()));
		}
		
		vmn->variable_ptr->remove_prior_samples();
	}
}

template <>
void setup_tree_gibbs_sampler(SubDiscreteFactorGraph & g, vector < vector <double> > & sum_marginals, vector < Potential * > & original_potentials)
{
	property_map<SubDiscreteFactorGraph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	
	graph_traits<SubDiscreteFactorGraph>::vertex_iterator u, u_end;
	graph_traits<SubDiscreteFactorGraph>::adjacency_iterator v, v_end;
	graph_traits<DiscreteFactorGraph>::vertex_descriptor w;
	subgraph<DiscreteFactorGraph>::children_iterator current_tree, tree_end;
	
	for ( tie(current_tree, tree_end) = g.children(); current_tree != tree_end; ++current_tree)
	{
		for (tie(u,u_end) = vertices (*current_tree); u!= u_end; ++u)
		{
			w = current_tree->local_to_global(*u);
			
			VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[w]);
			
			if ( vmn != NULL)
			{
				sum_marginals[get(vertex_index_map, w)] = vector < double> (vmn->variable_ptr->get_number_values(),0.0);
				
				original_potentials[ get(vertex_index_map, w)] = static_cast < DiscretePotential *> (g[w]->get_potential());
				
				ChainedSingleDiscretePotential * p = new ChainedSingleDiscretePotential(* vmn->variable_ptr);
				
				p->add_potential (static_cast < DiscretePotential *> (original_potentials[get(vertex_index_map, w)]));
				
				vmn->potential_ptr = p;
				
				for (tie(v, v_end)= adjacent_vertices ( w, g ); v != v_end; ++v)
				{
					if (! (current_tree->find_vertex(*v).second))
					{
						// Add that potential (Factor Graphs)
						
						p->add_potential  (static_cast < PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * > (g [*v])->potential_ptr);
					}
				}
				
			}
			
		}
		
	}
}

template <>
void remove_temp_observations(SubDiscreteFactorGraph & g)
{	
	graph_traits<SubDiscreteFactorGraph>::vertex_iterator u, u_end;
	graph_traits<SubDiscreteFactorGraph>::adjacency_iterator v, v_end;
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[*u]);
		
		if (vmn != NULL)
		{
			
			if (!vmn->variable_ptr->is_conditionned() )
			{
				vmn->potential_ptr->unset_observed_variable(* vmn->variable_ptr);
				
				for ( tie (v, v_end) = adjacent_vertices(*u,g); v != v_end; ++v)
				{
					static_cast < PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> (g[*v])->potential_ptr->unset_observed_variable(* vmn->variable_ptr);
				}
				
			}
		}
		
	}
}

template <>
void cleanup_tree_gibbs_sampler(DiscreteFactorGraph & g, vector < Potential * > & original_potentials)
{
	property_map<DiscreteFactorGraph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	graph_traits<DiscreteFactorGraph>::vertex_iterator u, u_end;
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{		
		VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[*u]);
		
		if (vmn != NULL)
		{
			
			delete vmn->potential_ptr;
			vmn->potential_ptr = static_cast <DiscretePotential *> (original_potentials[ get(vertex_index_map, *u)]);
		}
	}	
}

template <>
void set_temp_observations(SubDiscreteFactorGraph & g)
{	
	graph_traits<SubDiscreteFactorGraph>::vertex_iterator u, u_end;
	graph_traits<SubDiscreteFactorGraph>::adjacency_iterator v, v_end;
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[*u]);
		
		if (vmn != NULL)
		{
			vmn->potential_ptr->set_observed_variable(* vmn->variable_ptr);
			
			for ( tie (v, v_end) = adjacent_vertices(*u,g); v != v_end; ++v)
			{
				static_cast < PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> (g[*v])->potential_ptr->set_observed_variable(* vmn->variable_ptr);
			}
		}
		
	}	
}

template <>
void initialize_graph_variables_random(DiscreteFactorGraph & g)
{
	graph_traits<DiscreteFactorGraph>::vertex_iterator u, u_end;
	graph_traits<DiscreteFactorGraph>::adjacency_iterator v, v_end;
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> ( g[*u]);
		
		if (vmn != NULL)
		{
			vmn->variable_ptr->set_value_at_random();
			
			for ( tie (v, v_end) = adjacent_vertices(*u,g); v != v_end; ++v)
			{
				static_cast < PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<DiscreteFactorGraph>::vertex_descriptor> *> (g[*v])->potential_ptr->set_variable_value(* vmn->variable_ptr);
			}
			
		}
	}
	
}

//************** Visitor RecordParent Implementation *************//

template <class Graph>
RecordParent<Graph>::RecordParent ( Graph & a ) : g(a) 
{
}

template <class Graph>
template <class Edge>
void RecordParent<Graph>::operator()(Edge e, const Graph& )
{
	g[target(e,g)]->parent_node = source(e,g);
	
	//cout << "I recorded " << source(e,g) << " as the parent of " << target(e,g) << endl;
}	

//************** End of Visitor RecordParent Implementation *************//

//************** Visitor Collect Implementation *************//

template <class Graph>
Collect<Graph>::Collect(Graph & a): g(a)
{
}

template <class Graph>
template <class VertexDescriptor>
void Collect<Graph>::operator()(VertexDescriptor u, const Graph & )
{
	// Getting the parent of u
	
	VertexDescriptor v = g[u]->parent_node;
	
	if (u == v)
	{
		// Root Vertex. We don't do anything. (The root vertex is its own parent)
		return;
	}
	
	// u is the origin, v the destination.
	
	//cout << "Sending collect message from " << g[u]->index << " to " << g[v]->index << "." << endl;
	
	setup_message_computation (g, u , v);
	
	g[v]->receive_message ( g[u]->send_message( * g[v] ) );
	
}	

//************** End of Visitor Collect Implementation *************//

//************** Visitor Distribute Implementation *************//

template <class Graph>
Distribute<Graph>::Distribute(Graph & a): g(a)
{
}

template <class Graph>
template <class Vertex>
void Distribute<Graph>::operator()(Vertex v, const Graph& ) {
	
	// Be conscious of the fact, that u is still the origin and v the destination... (it is NOT switched from Collect in semantics, but in
	// terms of the graph algorithm IT IS...
	// What I mean is the visitor is called no longer on the origin but on the destination.
	
	Vertex u = g[v]->parent_node;
	
	if (u == v)
	{
		// Root Vertex. We don't do anything.
		return;
	}
	
	// u is the origin, v the destination.
	
	//cout << "Sending distribute message from " << g[u]->index << " to " << g[v]->index << "." << endl;
	
	setup_message_computation (g, u , v);
	
	g[v]->receive_message ( g[u]->send_message( * g[v] ) );
}

//************** End of Visitor Distribute Implementation *************//

//************** Visitor BackwardSampling Implementation *************//

template <class Graph>
BackwardSampling<Graph>::BackwardSampling(Graph & a): g(a)
{
}

template <class Graph>
template <class Vertex>
void BackwardSampling<Graph>::operator()(Vertex v, const Graph& )
{

	// Note that as it samples, the recorded samples are kept in memory
	
	Vertex u = g[v]->parent_node;
	
	//cout << "Sampling from: " << g[v]->index << endl; 
	
	if (u == v)
	{
		// Root Vertex. We sample from the marginal
		
		g[v]->sample_from_joint_root();
		
		return;
	}
	
	// We setup the message computation, this will be used even in the joint sampling part
	
	setup_message_computation(g, u, v);
	
	// This is the joint sampling part
	
	g[v]->sample_from_joint(* g[u]);
}

//************** End of Visitor BackwardSampling Implementation *************//

template <class Graph>
void initialize_graph_contents(Graph & g)
{
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		g[*u]->initialize();
	}
	
	remove_temp_observations(g);
}


template <class Graph>
void tree_gibbs_sampler(Graph & g, const unsigned int max_steps, const unsigned int burnoff_steps = 0, const bool rao_blackwell = true, const bool time = false)
{
	typename graph_traits<Graph>::vertex_iterator  u, u_end;
	typename graph_traits<Graph>::adjacency_iterator  v, v_end;
	typename subgraph<Graph>::children_iterator current_tree, tree_end;
	
	vector < vector < double> > sum_marginals (num_vertices(g));
	vector < Potential * > original_potentials (num_vertices(g));
	
	fibonnacci_number_generator.seed(std::time(NULL));
	
	initialize_graph_variables_random(g);
	
	// We now partition the graph. Warning: we send directly the subgraph type when we send to partition_graph_into_trees, not the graph
	
	//display_graph(g);
	
	subgraph <Graph> subgraph_g;
	copy_graph(g, subgraph_g);
	partition_graph_into_trees(subgraph_g);
	
	//display_subgraph(subgraph_g);
	
	// Here we setup stuff based on the partition. This is a specialized templated function.
	
	setup_tree_gibbs_sampler(subgraph_g, sum_marginals, original_potentials);
	
	for ( tie (current_tree, tree_end) = subgraph_g.children(); current_tree != tree_end; ++current_tree)
	{
		set_temp_observations(* current_tree);
	}
	
	
	cout << "Running MCMC Tree Gibbs Sampling with " << max_steps <<  (time ? " seconds."  : " steps." ) << endl;
	
	unsigned int current_gibbs_step (0);
	
	bool record = false;
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	double time_used(0.0);
	
	if (time)
	{
		global_timer.restart();
	}
	
	
	// Main loop after preparation stuff (Gibbs steps)
	
	while (time ? time_used < max_steps : current_gibbs_step < max_steps)
	{
		
		// Set up looping over trees (components over which we can infer probabilities)
		
		for ( tie (current_tree, tree_end) = subgraph_g.children(); current_tree != tree_end; ++current_tree)
		{
			//cout << "*************************************" << endl;
			
			// Initialization of the tree (we clear previous messages, etc...).
			
			initialize_graph_contents(* current_tree);
			
			// Must run inference
			
			belief_propagation_with_backward_sampling(* current_tree);
			
			//  display_variables_graph(*current_tree);
			
			// We have now obtained a sample from the joint. The following will also take care of RB estimates
			
			for (tie(u,u_end) = vertices (*current_tree); u!= u_end; ++u)
			{	
				// We could remove the following line to prevent use of RB, and only use Gibbs counting
				
				(*current_tree)[*u]->add_rao_blackwell_estimates(sum_marginals[current_tree->local_to_global(*u)]);
				
				(*current_tree)[*u]->set_variable_to_joint_sampled_value();
			}
			
			set_temp_observations((*current_tree));
		}
		
		time_used = global_timer.elapsed();
		++current_gibbs_step;
		
		if (!time)
		{
			if ( (current_gibbs_step % (max_steps/100) == 0) || (max_steps < 100) )
			{ 
				cout << "." << flush; 
			}
			
			if ( current_gibbs_step > burnoff_steps )
			{ 
				record = true; 
			}
			
			if (monitoring && record)
			{
				// Following is not portable...
				
				if (rao_blackwell)
				{
					for (tie (u, u_end) = vertices (g); u != u_end; ++u)
					{
						g[*u]->set_rao_blackwell_estimates(sum_marginals[*u]);
					}
				}
				else
				{
					for (tie (u, u_end) = vertices (g); u != u_end; ++u)
					{
						g[*u]->compute_marginals_from_samples();
					}
				}
				
				current_experiment->callback_monitoring(current_gibbs_step);
			}
			
			
		}
		
		else
		{
			if (time_used > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			if ( time_used > burnoff_steps )
			{ 
				record = true; 
			}
			
			
			if (monitoring && record)
			{
				// Following is not portable...
				
				if (rao_blackwell)
				{
					for (tie (u, u_end) = vertices (g); u != u_end; ++u)
					{
						g[*u]->set_rao_blackwell_estimates(sum_marginals[*u]);
					}
				}
				else
				{
					for (tie (u, u_end) = vertices (g); u != u_end; ++u)
					{
						g[*u]->compute_marginals_from_samples();
					}
				}
				
				current_experiment->callback_monitoring(time_used);
				
				
			}
			
			
			
		}
	}
	
	cout << endl;
	
	// Following is not portable to other types of graphs than adjacency_list...
	
	if (rao_blackwell)
	{
		for (tie (u, u_end) = vertices (g); u != u_end; ++u)
		{
			g[*u]->set_rao_blackwell_estimates(sum_marginals[*u]);
		}
	}
	
	else
	{
		for (tie (u, u_end) = vertices (g); u != u_end; ++u)
		{
			g[*u]->compute_marginals_from_samples();
		}
	}
	
	if (time)
	{
		cout << "We were able to generate " << current_gibbs_step << " Tree Gibbs samples." << endl;
	}
	
	// Deallocating memory
	
	cleanup_tree_gibbs_sampler(g, original_potentials);
}

// Implementation of a Simple Gibbs Sampler algorithm, which samples nodes one at a time.

template <class Graph>
void gibbs_sampler(Graph & g, const unsigned int max_steps, const unsigned int burnoff_steps, const bool time = false)
{
	typename graph_traits<Graph>::vertex_iterator  u, u_end;
	
	cout << "Running Gibbs Sampler with " << max_steps <<  (time ? " seconds and "  : " steps and " ) << burnoff_steps << " burnoff steps." << endl;
	
	assert ( burnoff_steps < max_steps );
	
	fibonnacci_number_generator.seed(std::time(NULL));
	
	initialize_graph_variables_random(g);
	
	// Now we setup the proposals, these should not change
	
	for ( tie (u, u_end) = vertices(g) ; u != u_end; ++u)
	{	
		// Here we do some graph dependent setup for setting the proposal
		
		setup_gibbs_proposal(g, *u);
	}
	
	unsigned int current_gibbs_step = 0;
	
	bool record = false;
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	double time_used(0.0);
	
	if (time)
	{
		global_timer.restart();
	}
	
	// Begin of the main Gibbs loop
	
	while (time ? time_used < max_steps : current_gibbs_step <= max_steps)
	{
		
		for ( tie (u, u_end) = vertices(g) ; u != u_end; ++u)
		{	
			// Here come the generic virtual method
			
			g[*u]->sample_from_proposal(record);
		}
		
		++current_gibbs_step;
		time_used = global_timer.elapsed();
		
		if (!time)
		{
			if ( (current_gibbs_step % (max_steps/100) == 0) || (max_steps < 100) )
			{ 
				cout << "." << flush; 
			}
			
			
			if (monitoring && record)
			{
				for (tie (u, u_end) = vertices (g); u != u_end; ++u)
				{
					g[*u]->compute_marginals_from_samples();
				}
				
				current_experiment->callback_monitoring(current_gibbs_step);
			}
			
			
			if ( current_gibbs_step > burnoff_steps )
			{ 
				record = true; 
			}
		}
		
		else
		{
			if (time_used > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			if (monitoring && record)
			{
				for (tie (u, u_end) = vertices (g); u != u_end; ++u)
				{
					g[*u]->compute_marginals_from_samples();
				}
				
				current_experiment->callback_monitoring(time_used);
			}
			
			if ( time_used > burnoff_steps )
			{ 
				record = true; 
			}
		}
		
	}
	
	// End of the Gibbs loop
	
	for (tie (u, u_end) = vertices (g); u != u_end; ++u)
	{
		g[*u]->compute_marginals_from_samples();
	}
	
	cout << endl;
	
	if (time)
	{
		cout << "We were able to generate " << current_gibbs_step << " Gibbs samples." << endl;
	}
}

// The message_passing algorithm consists in two depth first search passes.

template <class Graph>
void belief_propagation(Graph & g)
{	
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	
	// We record every node as its own parent
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->parent_node = *u;
	}
	
	depth_first_search(g, visitor(make_dfs_visitor(std::make_pair(Collect < Graph > (g),  RecordParent  < Graph > (g) ) ) ));	
	depth_first_search(g, visitor(make_dfs_visitor(Distribute< Graph > (g) ) ) );
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->compute_marginals_from_messages();
	}
}

template <class Graph>
void belief_propagation_with_backward_sampling(Graph & g)
{	
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	
	// We record every node as its own parent
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->parent_node = *u;
	}
	
	//cout << "Starting Collect" << endl;
	
	depth_first_search(g, visitor(make_dfs_visitor(std::make_pair(Collect < Graph > (g),  RecordParent  < Graph > (g) ) ) ));	
	
	//cout << "Starting B Sampling" << endl;
	
	depth_first_search(g, visitor(make_dfs_visitor(BackwardSampling< Graph > (g) ) ) );
	
	//cout << "Starting Distribute" << endl;
	
	remove_temp_observations(g);
	
	depth_first_search(g, visitor(make_dfs_visitor(Distribute< Graph > (g) ) ) );
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->compute_marginals_from_messages();
	}
	
	/*
	 Attempt at NOT using FFBS for debug purposes
	 
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->sample();
	}
	 */
}

template <class Graph>
void loopy_belief_propagation(Graph & g, const unsigned int max_steps, const bool time = false)
{		
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	typename graph_traits<Graph>::adjacency_iterator v, v_end;
	
	// Main loop, (loopy steps)
	
	cout << "Running Loopy Belief Propagation with " << max_steps <<  (time ? " seconds."  : " steps." ) << endl;
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	double time_used(0.0);
	
	unsigned int current_loopy_step(0);
	
	if (time)
	{
		global_timer.restart();
	}
	
	while (time ? time_used < max_steps : current_loopy_step < max_steps)
	{
		
		for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
		{
			for (tie (v, v_end) = adjacent_vertices(*u, g); v != v_end; ++v)
			{
				// Here we do some graph dependent setup for sending the message
				
				setup_message_computation (g, *u , *v);
				
				g[*v]->update_message ( g[*u]->send_message(* g[*v] ) );
			}
		}
		
		time_used = global_timer.elapsed();
		++current_loopy_step;
		
		if (!time)
		{
			if ( (current_loopy_step % (max_steps/100) == 0) || (max_steps < 100) )
			{ 
				cout << "." << flush; 
			}
			
			if (monitoring)
			{
				for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
				{
					g[*u]->compute_marginals_from_messages();
				}
				
				current_experiment->callback_monitoring(current_loopy_step);
			}
		}
		
		else
		{
			if (time_used > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			
			if (monitoring)
			{
				for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
				{
					g[*u]->compute_marginals_from_messages();
				}
				
				current_experiment->callback_monitoring(time_used);
			}
		}
		
	}
	
	// We are finished, let's get the marginals (beliefs)
	
	if (time)
	{
		cout << "\nWe were able to generate " << current_loopy_step << " Loopy steps." << endl;
	}
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->compute_marginals_from_messages();
	}
}


template <class Graph>
void forward_filtering_backward_sampling(Graph & g)
{	
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	
	// We record every node as its own parent
	
	for ( tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->parent_node = *u;
	}
	
	depth_first_search(g, visitor(make_dfs_visitor(std::make_pair(Collect < Graph > (g),  RecordParent  < Graph > (g) ) ) ));
	
	depth_first_search(g, visitor(make_dfs_visitor(BackwardSampling< Graph > (g) ) ) );
}
