#include <ParticleFilters.h>
#include <Potential.h>
#include <Node.h>
#include <RandomVariable.h>
#include <GraphTreePartition.h>
#include <OutputGraph.h>
#include <Experiment.h>

#include <ctime>
#include <boost/graph/subgraph.hpp>
#include <boost/graph/copy.hpp>

using namespace std;
using namespace boost;

// Explicit Template Instantiations //

typedef property< edge_index_t, unsigned int, ContinuousPotential *> NPEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, GraphicalModelNode *, NPEdgeProperty > NPGraph;

typedef property< edge_index_t, unsigned int, DiscretePotential *> DiscreteEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, MessageNode<DiscreteRandomVariable, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> *, DiscreteEdgeProperty > DiscreteGraph;

template void metropolis < NPGraph> (NPGraph &, const unsigned int, const unsigned int, bool);

template void non_parametric_tree_gibbs_sampler < NPGraph> (NPGraph &, const unsigned int, const unsigned int, const unsigned int, bool);

// End of Explicit Template instantiations //

//extern Experiment < DiscreteGraph> * current_discrete_experiment;

//******************* ParticleFilterProposal Implementation ****************//

ParticleFilterProposal::ParticleFilterProposal(): number_potentials(0), gaussian_proposal (fibonnacci_number_generator , normal_distribution <double> ())
{
}

double ParticleFilterProposal::get_proposal_weight(const double position) const
{
	//return (1/ sqrt( 2 * M_PI * sqrt(gaussian_proposal.distribution().sigma()) ) ) * exp( - ((position - gaussian_proposal.distribution().mean())*(position - gaussian_proposal.distribution().mean())) / (2*(gaussian_proposal.distribution().sigma())*(gaussian_proposal.distribution().sigma())  ) ); // 1/sqrt(2*pi*sigma) * exp(- (x-mean)2 / (2*sigma2) )
	
	// We can probably use the following formula ( proportional constant taken out)
	
	// cout << "dist mean " <<gaussian_proposal.distribution().mean();
	// cout <<  " position " << position;
	// cout << 
	
	return exp( - ( (position - gaussian_proposal.distribution().mean()) * (position - gaussian_proposal.distribution().mean() ) ) / (2*(gaussian_proposal.distribution().sigma())*(gaussian_proposal.distribution().sigma())  ) );
}


double ParticleFilterProposal::get_proposal_position(ContinuousRandomVariable & rv)
{
	//cout << "Gaussian sigma: " << gaussian_proposal.distribution().sigma() << endl;
	
	if (! rv.is_conditionned() )
	{
		rv.last_sampled_value = gaussian_proposal();
	}
	
	return rv.last_sampled_value;
}

void ParticleFilterProposal::setup_gaussian_proposal( ChainedSinglePotential * const ptr, const unsigned int variable_index)
{
	list < GaussianPotential * > new_lst;
	
	for (list <Potential * >::const_iterator it = ptr->potential_lst.begin(); it != ptr->potential_lst.end(); ++it)
	{
		GaussianPotential *  a = dynamic_cast<GaussianPotential*> (*it);
		
		if (a )
		{
			//cout << "Successfully downcasted" << endl;
			new_lst.push_back(a);
		}
		/*
		 else 
		 {
			 cout << " Not downcasted" << endl;	
		 }
		 */
	}
	
	setup_gaussian_proposal(new_lst, variable_index);
}


void ParticleFilterProposal::setup_gaussian_proposal(const list < GaussianPotential * >  & potential_lst, const unsigned int variable_index)
{
	// Assuming all of our Potentials are Gaussian
	
	// Must compute means
	
	number_potentials = potential_lst.size();
	means_sum = 0.0;
	
	if (number_potentials == 0)
	{
		return;
	}
	
	list <GaussianPotential * >::const_iterator it = potential_lst.begin();	
	constant_variance = (*it)->get_variance();
	
	for (list <GaussianPotential * >::const_iterator it = potential_lst.begin(); it != potential_lst.end(); ++it) 
	{
		means_sum += (*it)->get_variable_mean(variable_index);
		assert ( (*it)->get_variance() == constant_variance );
	}
	
	/*
	 if (variable_index == 2)
	 {
		 cout << " Number_potentials, means_sum, constant_variance/number_potentials: " << number_potentials << ", " << means_sum / number_potentials << ", " << constant_variance/number_potentials << endl;
	 }
	 */
	
	// Here we put directly the new proposal into effect
	
	gaussian_proposal.distribution() = normal_distribution < double > ( means_sum / number_potentials, sqrt (constant_variance/number_potentials) );
	
	//cout << "Gaussian mean: " << gaussian_proposal.distribution().mean() << endl;
}

void ParticleFilterProposal::add_gaussian_proposal(const double new_mean, const double new_variance)
{
	
	if (number_potentials != 0)
	{
		assert ( new_variance == constant_variance );
	}
	
	//cout << "New variance" << new_variance << endl;
	
	// Here we put directly the new proposal into effect
	
	// cout << "Proposal will be (number_potentials, new_mean, means_sum, constant_variance, means_sum/number_potentials): " <<  number_potentials+1 << ", " << new_mean << ", " << means_sum << ", " <<  constant_variance << ", " << (means_sum+new_mean) / (number_potentials+1) << endl;
	
	gaussian_proposal.distribution() = normal_distribution < double > ( (means_sum+new_mean) / (number_potentials+1), sqrt (new_variance/(number_potentials+1) ) );
}

void ParticleFilterProposal::set_gaussian_variance(const double new_variance)
{
	gaussian_proposal.distribution() = normal_distribution < double > ( gaussian_proposal.distribution().mean(), sqrt (new_variance) );
}

// Deprecated methods

/*
 void ParticleFilterProposal::setup_gaussian (const double a, const double b) // a is mean, b is variance
 {
	 //gaussian_proposal = variate_generator < lagged_fibonacci607 &, normal_distribution <double> >(fibonnacci_number_generator , normal_distribution < > ( a, b ));
	 
	 //normal_distribution <double> under_dist = gaussian_proposal.distribution() = normal_distribution < double > ( a, b );
	 
	 gaussian_proposal.distribution() = normal_distribution < double > ( a, sqrt (b) );
	 //mean = a;
	 //variance = b;
 }
 
 */

/*
 double ParticleFilterProposal::get_proposal_position()
 {	
	 cout << "Gaussian sigma: " << gaussian_proposal.distribution().sigma() << endl;
	 
	 return gaussian_proposal();
 }
 */

//******************* End of ParticleFilterProposal Implementation ****************//


//*************** DiscreteParticleFilterProposal Implementation ******************//

DiscreteParticleFilterProposal::DiscreteParticleFilterProposal()
{	
}

void DiscreteParticleFilterProposal::setup_discrete_proposal( const ChainedSingleDiscretePotential & csdp, DiscreteRandomVariable  & drv)
{
	general_sampling_probabilities.clear();
	general_sampling_probabilities.resize( drv.get_number_values() , 1.0);
	current_sampling_probabilities.clear();
	current_sampling_probabilities.resize( drv.get_number_values() , 1.0);
	
	//unsigned int variable_index = drv.get_index();
	
	for (unsigned int sp = 0 ; sp < general_sampling_probabilities.size(); ++sp)
	{
		for (list <DiscretePotential * >::const_iterator it = csdp.potential_lst.begin(); it != csdp.potential_lst.end(); ++it)
		{
			drv.set_value(sp);
			
			(*it)->set_variable_value (drv);
			
			general_sampling_probabilities[sp] = general_sampling_probabilities[sp] * (*it)->get_potential_value ();
		}
	}	
}

void DiscreteParticleFilterProposal::add_discrete_proposal(DiscretePotential & dp, DiscreteRandomVariable  & drv)
{
	//unsigned int variable_index = drv.get_index();
	
	for (unsigned int sp = 0 ; sp < current_sampling_probabilities.size(); ++sp)
	{
		drv.set_value(sp);
		dp.set_variable_value (drv);
		
		current_sampling_probabilities[sp] = general_sampling_probabilities[sp] * dp.get_potential_value ( );
	}	
}

unsigned int DiscreteParticleFilterProposal::get_proposal_position(DiscreteRandomVariable & rv)
{	
	if (!fast_double_normalize_over (current_sampling_probabilities.begin(), current_sampling_probabilities.end()))
	{
		cout << "Warning: all probabilities are 0" << endl;
	}
	
	return rv.sample_without_recording(current_sampling_probabilities);
}

inline double DiscreteParticleFilterProposal::get_proposal_weight(const unsigned int a) const
{
	return current_sampling_probabilities[a];	
}

//*************** End of DiscreteParticleFilterProposal Implementation ******************//

template <>
void initialize_particles <subgraph<DiscreteGraph>, unsigned int, DiscreteParticleFilterProposal> (subgraph<DiscreteGraph> & g, vector <unsigned int> & initial_positions, vector <double> & initial_weights, const graph_traits<DiscreteGraph>::vertex_descriptor root_node, DiscreteParticleFilterProposal & proposal)
{
	typedef graph_traits<DiscreteGraph>::vertex_descriptor VertexDescriptor;
	
	graph_traits<DiscreteGraph>::adjacency_iterator u, u_end;
	
	unsigned int number_particles = initial_weights.size();
	
	vector <double> cumulative_distribution (number_particles);
	vector <double> stratified_weights (number_particles);
	vector <unsigned int> original_particles (number_particles);
	
	Potential * internal_potential;
	DiscretePotential * observation_potential = NULL;
	
	internal_potential = static_cast < DiscretePotential *> ( g[root_node]->get_potential() );
	
	for (tie (u, u_end) = adjacent_vertices(root_node, g); u != u_end; ++u)
	{
		if (out_degree(*u,g) == 1)
		{
			observation_potential =  g[edge(root_node, *u, g).first];
		}
	}
	
	DiscreteRandomVariable * current_random_variable = g[root_node]->get_random_variable();
	
	// The proposal will be based on the values of the others nodes...
	
	proposal.setup_discrete_proposal(static_cast<ChainedSingleDiscretePotential &>( *internal_potential), *current_random_variable);
	
	vector <double>::iterator current_particle_weight = initial_weights.begin();
	
	for (vector <unsigned int>::iterator current_particle_position = initial_positions.begin(); current_particle_position != initial_positions.end(); ++current_particle_position)
	{
		*current_particle_position = proposal.get_proposal_position(*current_random_variable); // this also sets the current_random_variable.last_sampled_value
		
		// Compute the new weight
		
		observation_potential->set_variable_value (*current_random_variable);
		
		internal_potential->set_variable_value (*current_random_variable); // don't forget to set the internal potential also !!
		
		*current_particle_weight = ( observation_potential->get_potential_value() * internal_potential->get_potential_value() )/ proposal.get_proposal_weight(*current_particle_position);
		
		// cout << "Particle, weight of obs, internal, prop: "  << *current_particle_position << ", " << observation_potential->get_potential_value() << ", " <<internal_potential->get_potential_value() << ", " << proposal->get_proposal_weight(*current_particle_position) << endl;
		
		++current_particle_weight;
	}
	
	fast_double_normalize_over (initial_weights.begin(), initial_weights.end());
	
	resample (initial_positions, initial_weights, cumulative_distribution, stratified_weights, original_particles);
	
	//cout << "SIZE: " << initial_positions.size() << endl;
	
	//current_random_variable->set_mean_particles(initial_positions);
	
	//cout << "SIZE: " << initial_positions.size() << endl;
}

typedef iterator_property_map < vector< vector < unsigned int > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > DiscretePositionMap;
typedef iterator_property_map < vector< vector < double > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > DiscreteWeightMap;

template <>
graph_traits< subgraph<DiscreteGraph> >::vertex_descriptor forward_filtering_implementation < subgraph<DiscreteGraph>, DiscretePositionMap, DiscreteWeightMap > (subgraph<DiscreteGraph> & g, const graph_traits< subgraph<DiscreteGraph> >::vertex_descriptor root_node, DiscretePositionMap & positions, DiscreteWeightMap & weights)
{
	// Variable declarations
	
	typedef graph_traits<DiscreteGraph>::vertex_descriptor VertexDescriptor;
	
	DiscreteRandomVariable * current_random_variable, * previous_random_variable;
	
	unsigned int current_variable_index;
	
	VertexDescriptor double_previous_node, previous_node;
	
	VertexDescriptor current_node = VertexDescriptor();
	
	previous_node = root_node;
	
	double_previous_node = root_node;
	
	graph_traits<DiscreteGraph>::adjacency_iterator u, u_end;
	
	DiscreteParticleFilterProposal proposal;
	
	unsigned int number_particles = get (positions, root_node).size();
	
	initialize_particles (g, get (positions, root_node), get (weights, root_node), root_node, proposal);
	
	Potential * internal_potential = NULL;
	DiscretePotential * observation_potential = NULL;
	DiscretePotential * interaction_potential = NULL;
	
	vector <double> cumulative_distribution (number_particles);
	vector <double> stratified_weights (number_particles);
	vector <unsigned int> original_particles (number_particles);
	
	// We got our initial distribution covered, resample it ?? Currently, NO resampling is performed.
	
	// Initialization done.
	// Beginning of main loop.
	
	while (true)
	{
		
		// * We must obtain:
		
		// * potential coming from the edges linking the node to the other nodes not in the chain, call that internal_potential (because we arranged it as an internal potential)
		// * potential between node and observation, call that observation_potential;
		// * potential between the two nodes, call that interaction potential;
		
		for (tie (u, u_end) = adjacent_vertices(previous_node, g); u != u_end; ++u)
		{
			if (out_degree(*u,g) != 1)
			{
				
				if ( *u != double_previous_node)
				{
					current_node = *u;
					interaction_potential = g[ edge(previous_node, current_node, g).first];
					
					internal_potential = static_cast < DiscretePotential *> (g[current_node]->get_potential() );
				}
			}
		}
		
		for (tie (u, u_end) = adjacent_vertices(current_node, g); u != u_end; ++u)
		{
			if (out_degree(*u,g) == 1)
			{
				observation_potential =  g[edge(current_node, *u, g).first];
			}
		}
		
		//cout << "Current node is " << current_node << " (" << g[current_node]->get_random_variable()->get_index() << ")" << ", previous was " << previous_node << " (" << g[previous_node]->get_random_variable()->get_index() << ")" << endl;
		
		//cout << "Current node (FF) is " << g[current_node]->get_random_variable()->get_index()  << endl;
		
		//previous_variable_index = g[previous_node]->get_random_variable()->get_index();
		
		current_random_variable = g[current_node]->get_random_variable();
		previous_random_variable = g[previous_node]->get_random_variable();
		current_variable_index = current_random_variable->get_index();
		
		// Sampling particle from proposal
		
		vector <double> & current_weights = get (weights, current_node);		
		vector <double> & previous_weights = get (weights, previous_node);
		vector <unsigned int> & current_positions = get (positions, current_node);		
		vector <unsigned int> & previous_positions = get (positions, previous_node);
		
		vector <unsigned int>::iterator previous_particle_position = previous_positions.begin();
		vector <double>::iterator previous_particle_weight = previous_weights.begin();
		vector <double>::iterator current_particle_weight = current_weights.begin();
		
		//cout << "First weight (before update): " << *current_weights.begin() << endl;
		
		proposal.setup_discrete_proposal(static_cast<ChainedSingleDiscretePotential &> (*internal_potential), *current_random_variable);
		
		for (vector <unsigned int>::iterator current_particle_position = current_positions.begin(); current_particle_position != current_positions.end(); ++current_particle_position)
		{
			// Sampling the new particle
			
			previous_random_variable->set_value (*previous_particle_position);
			
			interaction_potential->set_variable_value( *previous_random_variable);
			
			proposal.add_discrete_proposal(*interaction_potential, *current_random_variable);
			
			*current_particle_position = proposal.get_proposal_position(*current_random_variable); // This also sets the current_random_variable.last_sampled_value
			
			// Compute the new weight
			
			observation_potential->set_variable_value (*current_random_variable); // All of these are discrete potentials	
			
			interaction_potential->set_variable_value (*current_random_variable);
			
			// Setting Internal Potential
			
			internal_potential->set_variable_value (*current_random_variable);
			
			// cout << "New particle, weight of old particle, obs, int, internal, prop: " << *current_particle_position << ", " << *previous_particle_weight << ", " << observation_potential->get_potential_value() << ", " << interaction_potential->get_potential_value() << ", " <<internal_potential->get_potential_value() << ", " << proposal->get_proposal_weight(*current_particle_position) << endl;
			
			*current_particle_weight = ( *previous_particle_weight * observation_potential->get_potential_value() * interaction_potential->get_potential_value() * internal_potential->get_potential_value() )/ proposal.get_proposal_weight(*current_particle_position);
			
			// cout << "Weight of new particle: " << *current_particle_weight << endl;
			
			++previous_particle_position;
			++previous_particle_weight;
			++current_particle_weight;
		}
		
		// Normalize the weights 
		
		if (!fast_double_normalize_over (current_weights.begin(), current_weights.end()))
		{
			cout << "Weights all to 0" << endl;
			
		}
		
		// Resample part
		
		resample (current_positions, current_weights, cumulative_distribution, stratified_weights, original_particles);
		
		//g[current_node]->get_random_variable()->set_mean_particles(current_positions);
		
		if (current_node == 54)
		{
			for ( vector <unsigned int>:: iterator it = current_positions.begin(); it != current_positions.end(); ++it)
			{
				//cout << *it << endl;
			}
		}
		
		if ( out_degree(current_node,g) == 2)
		{
			// We are finished
			
			return current_node;
		}
		
		double_previous_node = previous_node;
		previous_node = current_node;
		
	}
}


template < >
void backward_smoothing_implementation < subgraph<DiscreteGraph>, DiscretePositionMap, DiscreteWeightMap > (subgraph<DiscreteGraph> & g, const graph_traits< subgraph<DiscreteGraph> >::vertex_descriptor tail_node, DiscretePositionMap & positions, DiscreteWeightMap & weights, const unsigned int number_samples)
{
	
	//cout << "there" << endl;
	
	typedef graph_traits< subgraph<DiscreteGraph> >::vertex_descriptor VertexDescriptor;
	
	VertexDescriptor double_next_node, next_node, current_node;
	
	graph_traits<subgraph<DiscreteGraph> >::adjacency_iterator u, u_end;
	
	unsigned int next_sample_position, current_sample_position;
	
	RandomVariable * next_random_variable;
	RandomVariable * current_random_variable;
	
	vector <double> current_new_weights (get (weights, tail_node).size());
	
	for (unsigned int current_sample = 0; current_sample < number_samples; ++current_sample)
	{
		//cout << "current sample" << current_sample << endl;
		
		next_node = tail_node;
		double_next_node = tail_node;
		current_node = tail_node;
		
		next_random_variable = g[next_node]->get_random_variable();
		current_random_variable = g[current_node]->get_random_variable();
		
		current_sample_position = static_cast < DiscreteRandomVariable * > (current_random_variable)->sample_from_particles(get (positions, tail_node), get (weights, tail_node));
		
		next_sample_position = current_sample_position;
		
		g[current_node]->get_potential()->set_variable_value(* current_random_variable); // set the internal potential
		
		DiscretePotential * interaction_potential = NULL;
		DiscretePotential * internal_potential = NULL;
		
		while(true)
		{
			for (tie (u, u_end) = adjacent_vertices(next_node, g); u != u_end; ++u)
			{
				if (out_degree(*u,g) != 1)
				{
					// FIX_ME: done
					
					if ( *u != double_next_node)
					{
						current_node = *u;
						interaction_potential = g[edge(next_node, current_node, g).first];
						internal_potential = static_cast < DiscretePotential *> (g[current_node]->get_potential() );
					}
				}
			}
			
			
			//unsigned int current_variable_index =  g[current_node]->get_random_variable()->get_index();
			
			next_random_variable = g[next_node]->get_random_variable();
			current_random_variable = g[current_node]->get_random_variable();
			
			//cout << "Current node (BS) is " << current_variable_index << endl;
			
			// Here this is the only time we set up the interaction potential based on the sampled value
			
			interaction_potential->set_variable_value(* next_random_variable);
			
			vector <unsigned int> & current_positions = get (positions, current_node);
			vector <double> & current_weights = get (weights, current_node);
			
			vector <double>::iterator current_particle_weight = current_weights.begin();
			vector <double>::iterator current_particle_new_weight = current_new_weights.begin();
			
			for (vector <unsigned int>::iterator current_particle_position = current_positions.begin(); current_particle_position != current_positions.end(); ++current_particle_position)
			{
				static_cast <DiscreteRandomVariable *> (current_random_variable)->set_value(*current_particle_position);
				
				interaction_potential->set_variable_value( * current_random_variable);
				
				//cout << "Updated Weight (previous_weight, interaction_potential): " << *current_particle_weight * interaction_potential->get_potential_value() << " ( " << *current_particle_weight << ", " << interaction_potential->get_potential_value() << " ) " << endl;
				
				*current_particle_new_weight = *current_particle_weight * interaction_potential->get_potential_value();
				
				++current_particle_weight;
				++current_particle_new_weight;
			}
			
			/*
			 cout << "Updated Weights: " << endl;
			 
			 for ( vector <double>::iterator it = current_weights.begin(); it != current_weights.end(); ++it )
			 {
				 cout << *it << endl;
			 }
			 */
			
			// Normalize the weights (CRUCIAL HERE)
			
			fast_double_normalize_over (current_new_weights.begin(), current_new_weights.end());
			
			// CHECK_ME: not sure about the node... Checked, OK
			
			current_sample_position = static_cast <DiscreteRandomVariable * > (current_random_variable)->sample_from_particles(current_positions,  current_new_weights);
			next_sample_position = current_sample_position;
			
			vector <double>:: iterator jt = current_new_weights.begin();
			
			if ( current_random_variable->get_index() == 55 )
			{
				for ( vector <unsigned int>:: iterator it = current_positions.begin(); it != current_positions.end(); ++it)
				{
					
					//cout << *it << " " <<  *jt << endl;
					++jt;
				}
				
						//	cout << "*******************" << endl;
			}
			

			
			static_cast <DiscreteRandomVariable * >  (current_random_variable)->set_mean_bw_particles(current_positions,  current_new_weights);
			
			internal_potential->set_variable_value(*current_random_variable);
			
			if ( out_degree(current_node,g) == 2)
			{
				// We are finished
				
				break;
			}
			
			double_next_node = next_node;
			next_node = current_node;		
		}
	}
}

template <>
void non_parametric_tree_gibbs_sampler_implementation < DiscreteGraph> (DiscreteGraph & g, const unsigned int max_steps, const unsigned int number_particles, const unsigned int backward_samples, bool time)
{
	typedef graph_traits<DiscreteGraph>::vertex_iterator  VertexIterator;
	typedef graph_traits<DiscreteGraph>::adjacency_iterator  AdjacencyIterator;
	typedef graph_traits<DiscreteGraph>::vertex_descriptor  VertexDescriptor;
	typedef graph_traits<DiscreteGraph>::edge_descriptor EdgeDescriptor;
	
	typedef graph_traits< subgraph<DiscreteGraph> >::vertex_descriptor SubGraphVertexDescriptor;
	
	VertexIterator u, u_end;
	AdjacencyIterator v,v_end;
	VertexDescriptor w;
	
	subgraph<DiscreteGraph>::children_iterator current_tree, current_tree_end;
	
	subgraph <DiscreteGraph> subgraph_g;
	
	copy_graph(g, subgraph_g);
	
	// Use the following function if you are dealing with a MRF with same and pair number of rows / columns 
	
	partition_mrf_graph_into_chains(subgraph_g);
	
	// Else use the following if you are dealing with a single chain
	
	//partition_simple_chain(subgraph_g);
	
	//display_subgraph(subgraph_g);
	
	// Following block set up the correct internal potential for a variable given the edges linking it to the other variables (conditionned)
	
	
	list < vector < vector < double > > * > trees_weights_map;
	list < vector < vector < unsigned int > > * > trees_positions_map;
	
	for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
	{
		vector < double > particles_weights (number_particles);
		
		vector < vector < double > > * weights_vector = new  vector < vector < double > > ( num_vertices (*current_tree), particles_weights);
		
		trees_weights_map.push_back(weights_vector);
		
		vector < unsigned int > particles_positions (number_particles);
		
		vector < vector < unsigned int > > * positions_vector = new  vector < vector < unsigned int > > ( num_vertices (*current_tree), particles_positions);
		
		trees_positions_map.push_back(positions_vector);
		
		for (tie(u,u_end) = vertices (*current_tree); u!= u_end; ++u)
		{
			w = current_tree->local_to_global(*u);
			
			for (tie(v, v_end)= adjacent_vertices ( w, g ); v != v_end; ++v)
			{
				if (!current_tree->find_vertex(*v).second)
				{
					// Add that edge potential to the ChainedSinglePotential
					
					// BUG HERE, it is the potential of the EDGE !!!!!  ... Fixed
					
					static_cast < ChainedSingleDiscretePotential * > (g[w]->get_potential())->add_potential( g[edge(w,*v,g).first] );
				}
			}
			
		}
		
	}
	
	// End
	
	list < vector < vector < unsigned int > > * >::iterator current_tree_positions_map = trees_positions_map.begin();
	list < vector < vector < double > > * >::iterator current_tree_weights_map = trees_weights_map.begin();
	
	vector < iterator_property_map < vector< vector < unsigned int > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > > positions_property_maps_vec;
	vector < iterator_property_map < vector< vector < double > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > > weights_property_maps_vec;
	
	for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
	{		
		positions_property_maps_vec.push_back(make_iterator_property_map( (*current_tree_positions_map)->begin(), get(vertex_index, *current_tree)));
		
		weights_property_maps_vec.push_back(make_iterator_property_map( (*current_tree_weights_map)->begin(), get(vertex_index, *current_tree)));
		
		++current_tree_weights_map;
		++current_tree_positions_map;
	}
	
	cout << "Running NP Chain Gibbs (Discrete) with " << max_steps << (time ? " seconds and " : " steps and ") << number_particles << " particles." << endl;
	
	fibonnacci_number_generator.seed(std::time(NULL));
	
	unsigned int gibbs_step(0);
	
	double time_used(0.0);
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	if (time)
	{
		global_timer.restart();
	}
	
	while (time ? time_used < max_steps : gibbs_step < max_steps )
	{
		++gibbs_step;
		
		if (!time)
		{
			if ( (gibbs_step % (max_steps/100) == 0) || (max_steps<100) )
			{ 
				cout << "." << flush; 
			}
			
			current_experiment->callback_monitoring(gibbs_step);
		}
		
		else
		{
			if (time_used > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			current_experiment->callback_monitoring(time_used);
		}
		
		// Set up looping over trees (components over which we can infer probabilities). WARNING: HERE IT MUST BE SPECIALLY CRAFTED CHAINS !
		
		vector < iterator_property_map < vector< vector < double > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > >::iterator weights_property_maps_it = weights_property_maps_vec.begin();
		vector < iterator_property_map < vector< vector < unsigned int > >::iterator, property_map < subgraph<DiscreteGraph>, vertex_index_t>::type > >::iterator positions_property_maps_it = positions_property_maps_vec.begin();
		
		for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
		{
			SubGraphVertexDescriptor root_node = vertex(0, *current_tree);
			
			for ( tie(u, u_end) = vertices(*current_tree); u != u_end; ++u)
			{
				if (out_degree(*u,*current_tree) == 2) // we found the "root" of our chain 
				{
					root_node = *u;
					break;
				}
			}
			
			// We get a sample from the joint using FFBS with PFs
			
			SubGraphVertexDescriptor tail_node = forward_filtering_implementation (*current_tree,  root_node, *positions_property_maps_it, *weights_property_maps_it);
			
			// This gives us the weights, and the particles, now run the backward smoothing part
			
			backward_smoothing_implementation (*current_tree,  tail_node, *positions_property_maps_it, *weights_property_maps_it, backward_samples);
			
			//  display_variables_graph(*current_tree);
			
			//++current_tree_weights_map;
			//++current_tree_positions_map;
			
			++positions_property_maps_it;
			++weights_property_maps_it;
		}
		
		// cout << "*********************************" << endl;
		
		time_used = global_timer.elapsed();
	}
	
	if (time)
	{
		cout << "We were able to generate " << gibbs_step << " complete tree-Gibbs samples." << endl;
	}
	
	// Here the Gibbs loop is done. That's all.
	
	// Need to clean up memory !!
}


template <>
void non_parametric_tree_gibbs_sampler < DiscreteGraph> (DiscreteGraph & g,  const unsigned int gibbs_steps, const unsigned int number_particles, const unsigned int backward_samples, bool time)
{	
	//typedef typename vertex_bundle_type<Graph>::type VertexClassPointer;
	//typedef typename iterator_traits<VertexClassPointer>::value_type VertexClass;
	
	property_map<DiscreteGraph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	
	typedef graph_traits<DiscreteGraph>::vertex_iterator  VertexIterator;
	
	VertexIterator u, u_end;
	
	vector < Potential * > original_potentials (num_vertices(g));
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		if ( out_degree(*u,g) != 1)
		{
			original_potentials[ get(vertex_index_map, *u)] = g[*u]->get_potential();
			
			ChainedSingleDiscretePotential * p = new ChainedSingleDiscretePotential (* g[*u]->get_random_variable());
			
			p->add_potential(static_cast < DiscretePotential *> ( g[*u]->get_potential() ) );
			
			g[*u]->set_potential(p);
		}
	}
	
	non_parametric_tree_gibbs_sampler_implementation ( g,  gibbs_steps, number_particles, backward_samples, time);
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		if ( out_degree(*u,g) != 1)
		{
			delete g[*u]->get_potential();
			g[*u]->get_random_variable()->obtain_inference_from_prior_samples();
			g[*u]->set_potential( original_potentials[ get(vertex_index_map, *u)]);
		}
	}
	
}

// No specialization needed for non_parametric_tree_gibbs_sampler_implementation (Graph & g, const unsigned int max_steps, const unsigned int number_particles);

//*********************** Generic Implementation ******************************//

//extern Experiment < NPGraph> * current_experiment;

template <class Graph>
void non_parametric_tree_gibbs_sampler(Graph & g, const unsigned int gibbs_steps, const unsigned int number_particles, const unsigned int backward_samples = 1, bool time = false)
{	
	//typedef typename vertex_bundle_type<Graph>::type VertexClassPointer;
	//typedef typename iterator_traits<VertexClassPointer>::value_type VertexClass;
	
	typename property_map<Graph, vertex_index_t>::type vertex_index_map = get(vertex_index, g);
	
	typedef typename graph_traits<Graph>::vertex_iterator  VertexIterator;
	
	VertexIterator u, u_end;
	
	vector < Potential * > original_potentials (num_vertices(g));
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{
		if ( out_degree(*u,g) != 1)
		{
			original_potentials[ get(vertex_index_map, *u)] = g[*u]->get_potential();
			
			ChainedSinglePotential * p = new ChainedSinglePotential ( g[*u]->get_random_variable()->get_index()); //g[*u]->get_potential()->clone();
			
			p->add_potential(g[*u]->get_potential() );
			
			g[*u]->set_potential(p);
		}
	}
	
	non_parametric_tree_gibbs_sampler_implementation ( g, gibbs_steps, number_particles, backward_samples, time);
	
	for (tie(u,u_end) = vertices (g); u!= u_end; ++u)
	{	
		if ( out_degree(*u,g) != 1)
		{
			delete g[*u]->get_potential();
			g[*u]->get_random_variable()->obtain_inference_from_prior_samples();
			g[*u]->set_potential( original_potentials[ get(vertex_index_map, *u)]);
		}
	}
	
}

template <class Graph>
void non_parametric_tree_gibbs_sampler_implementation(Graph & g, const unsigned int max_steps, const unsigned int number_particles, const unsigned int backward_samples, bool time)
{
	typedef typename graph_traits< Graph >::adjacency_iterator  BaseAdjacencyIterator;
	typedef typename graph_traits< Graph >::vertex_descriptor  BaseVertexDescriptor;
	
	typedef typename graph_traits< subgraph<Graph> >::vertex_iterator  SubGraphVertexIterator;
	typedef typename graph_traits< subgraph<Graph> >::vertex_descriptor  SubGraphVertexDescriptor;
	
	// typedef typename graph_traits<subgraph<Graph> >::edge_descriptor BaseEdgeDescriptor;
	
	typedef typename graph_traits<subgraph<Graph> >::edge_iterator SubGraphEdgeIterator;
	
	SubGraphVertexIterator u, u_end;
	BaseAdjacencyIterator v,v_end;
	BaseVertexDescriptor w;
	
	typename subgraph<Graph>::children_iterator current_tree, current_tree_end;
	
	subgraph <Graph> subgraph_g;
	
	copy_graph(g, subgraph_g);
	
	// If it is a simple chain, do this (no partitioning needed)
	
	//partition_simple_chain(subgraph_g);
	
	// Else do this for a MRF (with a PAIR and EQUAL number of rows / columns)
	
	partition_mrf_graph_into_chains(subgraph_g);
	
	//display_subgraph(subgraph_g);
	
	// Following block set up the correct internal potential for a variable given the edges linking it to the other variables (conditionned)
	
	list < vector < vector < double > > * > trees_weights_map;
	list < vector < vector < double > > * > trees_positions_map;
	
	for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
	{
		vector < double > particles_weights (number_particles);
		
		vector < vector < double > > * weights_vector = new  vector < vector < double > > ( num_vertices (*current_tree), particles_weights);
		
		trees_weights_map.push_back(weights_vector);
		
		vector < double > particles_positions (number_particles);
		
		vector < vector < double > > * positions_vector = new  vector < vector < double > > ( num_vertices (*current_tree), particles_positions);
		
		trees_positions_map.push_back(positions_vector);
		
		for (tie(u,u_end) = vertices (*current_tree); u!= u_end; ++u)
		{
			w = current_tree->local_to_global(*u);
			
			for (tie(v, v_end)= adjacent_vertices ( w, g ); v != v_end; ++v)
			{
				if (!current_tree->find_vertex(*v).second)
				{
					// Add that edge potential to the ChainedSinglePotential
					
					static_cast < ChainedSinglePotential * > (g[w]->get_potential())->add_potential( g[edge(w,*v,g).first] );
				}
			}
			
		}
		
	}
	
	// End
	
	
	list < vector < vector < double > > * >::iterator current_tree_weights_map = trees_weights_map.begin();
	list < vector < vector < double > > * >::iterator current_tree_positions_map = trees_positions_map.begin();
	
	vector < iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > > weights_property_maps_vec;
	vector < iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > > positions_property_maps_vec;
	
	
	for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
	{
		//iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > position_map = make_iterator_property_map( (*current_tree_positions_map)->begin(), get(vertex_index, *current_tree));
		
		//iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > weight_map = make_iterator_property_map( (*current_tree_weights_map)->begin(), get(vertex_index, *current_tree));
		
		positions_property_maps_vec.push_back(make_iterator_property_map( (*current_tree_positions_map)->begin(), get(vertex_index, *current_tree)));
		
		weights_property_maps_vec.push_back(make_iterator_property_map( (*current_tree_weights_map)->begin(), get(vertex_index, *current_tree)));
		
		++current_tree_weights_map;
		++current_tree_positions_map;
	}
	
	cout << "Running NP Chain Gibbs with " << max_steps << (time ? " seconds and " : " steps and ") << number_particles << " particles." << endl;
	
	// cout << "MapSize: " << trees_weights_map.size() << ", " << trees_positions_map.size() << endl;
	
	// Main loop after preparation stuff (Gibbs steps)
	
	fibonnacci_number_generator.seed(std::time(NULL));
	
	unsigned int gibbs_step(0);
	
	double time_used(0.0);
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	if (time)
	{
		global_timer.restart();
	}
	
	while (time ? time_used < max_steps : gibbs_step < max_steps )
	{
		++gibbs_step;
		
		if (!time)
		{
			if ( (gibbs_step % (max_steps/100) == 0) || (max_steps<100) )
			{ 
				cout << "." << flush; 
			}
		}
		
		else
		{
			if (time_used > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			current_experiment->callback_monitoring(time_used);
		}
		
		// Set up looping over trees (components over which we can infer probabilities). WARNING: HERE IT MUST BE SPECIALLY CRAFTED CHAINS !
		
		typename vector < iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > >::iterator weights_property_maps_it = weights_property_maps_vec.begin();
		typename vector < iterator_property_map < vector< vector < double > >::iterator, typename property_map < subgraph<Graph>, vertex_index_t>::type > >::iterator positions_property_maps_it = positions_property_maps_vec.begin();
		
		for ( tie(current_tree, current_tree_end) = subgraph_g.children(); current_tree != current_tree_end; ++current_tree)
		{
			/*
			 SubGraphEdgeIterator e, e_end;
			 
			 tie (e,e_end) = edges(g);
			 
			 for (tie (e,e_end) = edges(g); e != e_end; ++e)
			 {
				 g[*e]->debug_display_state();
			 }
			 */
			
			SubGraphVertexDescriptor root_node = vertex(0, *current_tree);
			
			for ( tie(u, u_end) = vertices(*current_tree); u != u_end; ++u)
			{
				if (out_degree(*u,*current_tree) == 2) // we found the "root" of our chain 
				{
					root_node = *u;
					break;
				}
			}
			
			//cout << "Root node of FF is: " << g[root_node]->get_random_variable()->get_index() << endl;
			
			SubGraphVertexDescriptor tail_node = forward_filtering_implementation (*current_tree,  root_node, *positions_property_maps_it, *weights_property_maps_it);
			
			// This gives us the weights, and the particles, now run the backward smoothing part
			
			backward_smoothing_implementation (*current_tree,  tail_node, *positions_property_maps_it, *weights_property_maps_it, backward_samples);
			
			//  display_variables_graph(*current_tree);
			
			//++current_tree_weights_map;
			//++current_tree_positions_map;
			
			++positions_property_maps_it;
			++weights_property_maps_it;
		}
		
		// cout << "*********************************" << endl;
		
		time_used = global_timer.elapsed();
	}
	
	cout << endl;
	
	if (time)
	{
		cout << "We were able to generate " << gibbs_step << " complete tree-Gibbs samples." << endl;
	}
	
	// Here the Gibbs loop is done. That's all.
}

template < class Graph, class PositionMap, class WeightMap>
typename graph_traits< Graph >::vertex_descriptor forward_filtering_implementation (Graph & g, const typename graph_traits< Graph >::vertex_descriptor root_node, PositionMap & positions, WeightMap & weights)
{
	// Variable declarations
	
	typedef typename graph_traits< Graph >::vertex_descriptor VertexDescriptor;
	
	ContinuousRandomVariable * current_random_variable;
	
	unsigned int current_variable_index, previous_variable_index ;
	
	VertexDescriptor double_previous_node, previous_node;
	
	VertexDescriptor current_node = VertexDescriptor ();
	
	previous_node = root_node;
	
	double_previous_node = root_node;
	
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	
	ParticleFilterProposal proposal;
	
	unsigned int number_particles = get (positions, root_node).size();
	
	initialize_particles (g, get (positions, root_node), get (weights, root_node), root_node, proposal);
	
	ContinuousPotential * internal_potential = NULL;
	ContinuousPotential * observation_potential = NULL;
	GaussianPotential * interaction_potential = NULL;
	
	vector <double> cumulative_distribution (number_particles);
	vector <double> stratified_weights (number_particles);
	vector <double> original_particles (number_particles);
	
	// We got our initial distribution covered, resample it ?? Currently, NO resampling is performed.
	
	// Initialization done.
	// Beginning of main loop.
	
	while (true)
	{
		// * We must obtain:
		
		// * potential coming from the edges linking the node to the other nodes not in the chain, call that internal_potential (because we arranged it as an internal potential)
		// * potential between node and observation, call that observation_potential;
		// * potential between the two nodes, call that interaction potential;
		
		for (tie (u, u_end) = adjacent_vertices(previous_node, g); u != u_end; ++u)
		{
			if (out_degree(*u,g) != 1)
			{
				
				if ( *u != double_previous_node)
				{
					current_node = *u;
					interaction_potential = static_cast <GaussianPotential *> (g[edge(previous_node, current_node, g).first]);
					
					// WARNING HERE: NOT SURE about the legitimacy of the cast, ConstantPotential should inherit both from Continuous and DiscretePotential
					
					internal_potential = static_cast <ContinuousPotential *>(g[current_node]->get_potential());
				}
			}
		}
		
		for (tie (u, u_end) = adjacent_vertices(current_node, g); u != u_end; ++u)
		{
			if (out_degree(*u,g) == 1)
			{
				observation_potential =  g[edge(current_node, *u, g).first];
				//cout << "Observation node is " << *u << " (" << g[*u]->get_random_variable()->get_index() << ")" << endl;
			}
		}
		
		//cout << "Current node (FF) is " << g[current_node]->get_random_variable()->get_index()  << endl;
		
		previous_variable_index = g[previous_node]->get_random_variable()->get_index();
		
		current_random_variable = static_cast<ContinuousRandomVariable *> (g[current_node]->get_random_variable());
		current_variable_index = current_random_variable->get_index();
		
		// Sampling particle from proposal
		
		vector <double> & current_weights = get (weights, current_node);		
		vector <double> & previous_weights = get (weights, previous_node);
		vector <double> & current_positions = get (positions, current_node);		
		vector <double> & previous_positions = get (positions, previous_node);
		
		vector <double>::iterator previous_particle_position = previous_positions.begin();
		vector <double>::iterator previous_particle_weight = previous_weights.begin();
		vector <double>::iterator current_particle_weight = current_weights.begin();
		
		//cout << "First weight (before update): " << *current_weights.begin() << endl;
		
		proposal.setup_gaussian_proposal(static_cast<ChainedSinglePotential *>(internal_potential), current_variable_index);
		
		for (vector <double>::iterator current_particle_position = current_positions.begin(); current_particle_position != current_positions.end(); ++current_particle_position)
		{
			// Sampling the new particle
			
			// cout << "Particle number " << current_particle_position - current_positions.begin() << endl;
			
			interaction_potential->set_variable_value (previous_variable_index, *previous_particle_position);
			
			//cout << "Var:" << interaction_potential->get_variance() << endl;
			
			proposal.add_gaussian_proposal(*previous_particle_position, interaction_potential->get_variance());
			
			*current_particle_position = proposal.get_proposal_position(*current_random_variable); // This also sets the current_random_variable.last_sampled_value
			
			// Compute the new weight
			
			observation_potential->set_variable_value (*current_random_variable); // It is the Bernouilli one, last_sampled_value must be set from the particle. done now (hope)
			
			interaction_potential->set_variable_value (*current_random_variable); // It is a Gaussian potential
			
			// Setting Internal Potential
			
			internal_potential->set_variable_value (*current_random_variable); // don't forget to set the internal potential also !!
			
			*current_particle_weight = ( *previous_particle_weight * observation_potential->get_potential_value() * interaction_potential->get_potential_value() * internal_potential->get_potential_value() )/ proposal.get_proposal_weight(*current_particle_position);
			
			/*
			 if (*current_particle_weight == 0.0)
			 {
				 cout << "New particle, proposal mean, proposal variance: " << *current_particle_position << " , " <<  proposal.debug_get_gaussian_mean() << " , " << proposal.debug_get_gaussian_variance() << endl;
				 
				 cout << "Observation, interaction, internal, proposal: "  << observation_potential->get_potential_value() << " , " << interaction_potential->get_potential_value() << ", " <<internal_potential->get_potential_value() << " , " << proposal.get_proposal_weight(*current_particle_position) << endl;
			 }
			 */
			
			// cout << "Weight of new particle: " << *current_particle_weight << endl;
			
			++previous_particle_position;
			++previous_particle_weight;
			++current_particle_weight;
		}
		
		// Normalize the weights 
		
		/*
		 cout << "Weights: " << endl;
		 
		 for ( vector <double>::iterator it = current_weights.begin(); it != current_weights.end(); ++it )
		 {
			 cout << *it << endl;
		 }		
		 
		 cout << "done" << endl << endl << endl;
		 
		 vector <double> test_weights = get (weights, current_node);
		 
		 for ( vector <double>::iterator it = test_weights.begin(); it != test_weights.end(); ++it )
		 {
			 cout << *it << endl;
		 }	
		 
		 cout << "done" << endl << endl << endl;
		 */
		
		if ( !fast_double_normalize_over (current_weights.begin(), current_weights.end()) )
		{
			cout << "Weights equal 0.0" << endl;
			
			for ( vector <double>::iterator it = current_weights.begin(); it != current_weights.end(); ++it )
			{
				*it = 1.0;
			}	
			
		}
		
		// Resample part
		
		resample ( current_positions, current_weights, cumulative_distribution, stratified_weights, original_particles);
		
		// End of one step of the forward pass
		
		/*
		 cout << "Particles: " << endl;
		 
		 for ( vector <double>::iterator it = current_positions.begin(); it != current_positions.end(); ++it )
		 {
			 cout << *it << endl;
		 }
		 */
		
		if ( out_degree(current_node,g) == 2)
		{
			// We are finished
			
			/*
			 cout << "Particles (at the last step): " << endl;
			 
			 for ( vector <double>::iterator it = current_positions.begin(); it != current_positions.end(); ++it )
			 {
				 cout << *it << endl;
			 }
			 */
			return current_node;
		}
		
		double_previous_node = previous_node;
		previous_node = current_node;
		
	}
	
	//delete proposal;
	
}

// We are going to try to implement a way of getting more samples in the backward pass

template < class Graph, class PositionMap , class WeightMap>
void backward_smoothing_implementation (Graph & g, const typename graph_traits<Graph>::vertex_descriptor tail_node, PositionMap & positions, WeightMap & weights, const unsigned int number_samples)
{
	typedef typename graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	VertexDescriptor double_next_node, next_node, current_node;
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	
	double next_sample_position, current_sample_position;
	
	RandomVariable * next_random_variable;
	RandomVariable * current_random_variable;
	
	vector <double> current_new_weights (get (weights, tail_node).size());
	
	for (unsigned int current_sample = 0; current_sample < number_samples; ++current_sample)
	{
		
		next_node = tail_node;
		double_next_node = tail_node;
		current_node = tail_node;
		
		next_random_variable = g[next_node]->get_random_variable();
		current_random_variable = g[current_node]->get_random_variable();
		
		// We sample the tail node variable, note that we could normalize weights there...
		
		current_sample_position = static_cast < ContinuousRandomVariable * > (current_random_variable)->sample_from_particles(get (positions, tail_node), get (weights, tail_node));
		
		next_sample_position = current_sample_position;
		
		g[current_node]->get_potential()->set_variable_value(* current_random_variable); // set the internal potential
		
		GaussianPotential * interaction_potential = NULL;
		ContinuousPotential * internal_potential = NULL;
		
		while (true)
		{
			for (tie (u, u_end) = adjacent_vertices(next_node, g); u != u_end; ++u)
			{
				if (out_degree(*u,g) != 1)
				{
					// FIX_ME: done
					
					if (*u != double_next_node)
					{
						current_node = *u;
						interaction_potential = static_cast <GaussianPotential *> (g[edge(next_node, current_node, g).first]);
						internal_potential = static_cast <ContinuousPotential *> (g[current_node]->get_potential() );
					}
				}
			}
			
			next_random_variable = g[next_node]->get_random_variable();
			
			current_random_variable = g[current_node]->get_random_variable();
			unsigned int current_variable_index =  current_random_variable->get_index();
			
			//cout << "Current node (BS) is " << current_variable_index << endl;
			
			// Here this is the only time we set up the potentials based on the sampled value
			
			interaction_potential->set_variable_value(* next_random_variable);
			
			vector <double> & current_positions = get (positions, current_node);
			vector <double> & current_weights = get (weights, current_node);
			
			vector <double>::iterator current_particle_weight = current_weights.begin();
			vector <double>::iterator current_particle_new_weight = current_new_weights.begin();
			
			for (vector <double>::iterator current_particle_position = current_positions.begin(); current_particle_position != current_positions.end(); ++current_particle_position)
			{
				interaction_potential->set_variable_value( current_variable_index, *current_particle_position);
				
				//cout << "Updated Weight (previous_weight, interaction_potential): " << *current_particle_weight * interaction_potential->get_potential_value() << " ( " << *current_particle_weight << ", " << interaction_potential->get_potential_value() << " ) " << endl;
				
				*current_particle_new_weight = *current_particle_weight * interaction_potential->get_potential_value();
				
				++current_particle_weight;
				++current_particle_new_weight;
			}
			
			/*
			 cout << "Updated Weights: " << endl;
			 
			 for ( vector <double>::iterator it = current_weights.begin(); it != current_weights.end(); ++it )
			 {
				 cout << *it << endl;
			 }
			 */
			
			// Normalize the weights (CRUCIAL HERE)
			
			fast_double_normalize_over (current_new_weights.begin(), current_new_weights.end());
			
			// We now sample the current RV, with respect to the updated weights
			
			current_sample_position = static_cast <ContinuousRandomVariable *> (current_random_variable)->sample_from_particles(current_positions,  current_new_weights);
			next_sample_position = current_sample_position;
			
			// And we set the internal Potential based on the value obtained
			
			internal_potential->set_variable_value(* current_random_variable);
			
			static_cast <ContinuousRandomVariable * >  (current_random_variable)->set_mean_bw_particles(current_positions,  current_new_weights);
			
			if ( out_degree(current_node,g) == 2)
			{
				// We are finished
				
				break;
			}
			
			double_next_node = next_node;
			next_node = current_node;		
		}
		
	}
	
	//dynamic_cast< ContinuousRandomVariable *> (g[vertex(0,g)]->get_random_variable())->debug_display_state();
}

template < class Graph, class PositionType, class ParticleFilterProposalType >
void initialize_particles (Graph & g, vector <PositionType> & initial_positions, vector <double> & initial_weights, typename graph_traits<Graph>::vertex_descriptor root_node, ParticleFilterProposalType & proposal)
{
	typedef typename graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	
	vector <double>::iterator current_particle_weight;
	
	ContinuousPotential * internal_potential = NULL;
	ContinuousPotential * observation_potential = NULL;
	
	internal_potential = static_cast < ContinuousPotential * > (g[root_node]->get_potential());
	
	for (tie (u, u_end) = adjacent_vertices(root_node, g); u != u_end; ++u)
	{
		if (out_degree(*u,g) == 1)
		{
			observation_potential =  g[edge(root_node, *u, g).first];
		}
	}
	
	ContinuousRandomVariable * current_random_variable = static_cast<ContinuousRandomVariable *> (g[root_node]->get_random_variable());
	unsigned int current_variable_index (current_random_variable->get_index());
	
	// The proposal will be based on the values of the others nodes...
	
	proposal.setup_gaussian_proposal(static_cast<ChainedSinglePotential *>(internal_potential), current_variable_index);
	
	while ( true )
	{
		current_particle_weight = initial_weights.begin();
		
		for (vector <double>::iterator current_particle_position = initial_positions.begin(); current_particle_position != initial_positions.end(); ++current_particle_position)
		{
			*current_particle_position = proposal.get_proposal_position(*current_random_variable); // this also sets the current_random_variable.last_sampled_value
			
			// Compute the new weight
			
			observation_potential->set_variable_value (*current_random_variable); // this is the Bernouilli
			
			internal_potential->set_variable_value (*current_random_variable); // don't forget to set the internal potential also !!
			
			//cout << "Current particle, Observation, Internal: " << *current_particle_position << ", " << observation_potential->get_potential_value() << ", " << internal_potential->get_potential_value() << endl;
			
			//cout << "Current particle weight: " << *current_particle_weight << endl;
			
			//cout << "Prop: " << proposal.get_proposal_weight(*current_particle_position) << endl;
			
			*current_particle_weight = ( observation_potential->get_potential_value() * internal_potential->get_potential_value() )/ proposal.get_proposal_weight(*current_particle_position);
			
			// cout << "Particle, weight of obs, internal, prop: "  << *current_particle_position << ", " << observation_potential->get_potential_value() << ", " <<internal_potential->get_potential_value() << ", " << proposal->get_proposal_weight(*current_particle_position) << endl;
			
			++current_particle_weight;
		}
		
		
		if ( accumulate(initial_weights.begin(), initial_weights.end(), 0.0) != 0.0 )
		{
			return;
		}
		
		cout << "The sum of weights equal 0, we increase variance." << endl;
		
		proposal.set_gaussian_variance(10.0*proposal.debug_get_gaussian_variance());
		
	}
	
}

template <class PositionType >
void resample( vector < PositionType > & current_positions, vector < double> & current_weights, vector < double> & cumulative_distribution, vector < double> & stratified_weights, vector < PositionType > & original_particles)
{	
	// Compute cumulative distribution
	
	double sum = 0.0;
	
	vector <double>::iterator current_particle_weight = current_weights.begin();
	
	unsigned int number_particles = current_weights.size();
	
	for ( vector <double>::iterator it = cumulative_distribution.begin(); it != cumulative_distribution.end(); ++it )
	{
		sum += *current_particle_weight;
		*it = sum;
		++current_particle_weight;
	}
	
	sum = uniform_distribution_0_1()/number_particles;
	
	current_particle_weight = current_weights.begin();
	
	for ( vector <double>::iterator it = stratified_weights.begin(); it != stratified_weights.end(); ++it )
	{
		*it = sum;
		sum += 1.0/number_particles;
	}
	
	copy(current_positions.begin(), current_positions.end(), original_particles.begin());
	
	vector <double>::iterator cumulative_distribution_it;
	typename vector < PositionType >::iterator selected_particle;
	vector <double>::iterator stratified_weights_it = stratified_weights.begin();
	
	selected_particle = original_particles.begin();
	cumulative_distribution_it = cumulative_distribution.begin();
	
	for (typename vector < PositionType >::iterator current_particle_position = current_positions.begin(); current_particle_position != current_positions.end(); ++current_particle_position)
	{
		while ( *stratified_weights_it > *cumulative_distribution_it)
		{
			++selected_particle;
			++cumulative_distribution_it;
		}
		
		*current_particle_position = *selected_particle;
		
		++stratified_weights_it;
	}
	
	// I think this is all for the resampling, set the weights to 1/number_particles
	
	for ( vector <double>::iterator it = current_weights.begin(); it != current_weights.end(); ++it )
	{
		*it = 1.0/number_particles;
	}
	
}

//*********** Other NP algorithms **************//

// This in fact is just a gibbs sampler with embedded Metropolis-Hastings steps.

template <class Graph>
void metropolis (Graph & g, const unsigned int max_steps, const unsigned int burnoff_steps = 0, bool time = false)
{
	vector < pair < double, double > > samples_vector ( num_vertices (g), pair < double, double > (0.0, 0.0) );
	
	metropolis_implementation(g, make_iterator_property_map(samples_vector.begin(), get(vertex_index,g)), max_steps, burnoff_steps, time);
}

template <class Graph, class SampleMap>
void metropolis_implementation (Graph & g, SampleMap previous_samples, const unsigned int max_steps, const unsigned int burnoff_steps, bool time)
{
	typename graph_traits<Graph>::vertex_iterator u, u_end;
	typename graph_traits<Graph>::adjacency_iterator v, v_end;
	
	ContinuousRandomVariable * current_random_variable;
	unsigned int current_variable_index;
	
	double new_observation_weight, new_sample;
	
	ContinuousPotential * observation_potential = NULL;
	
	ParticleFilterProposal proposal;
	
	// Seeding the RNG with a truely random value, should be replaced by clock() or something
	
	//fibonnacci_number_generator.seed(66);
	
	bool record = false;
	
	assert ( burnoff_steps < max_steps);
	
	fibonnacci_number_generator.seed(std::time(NULL));
	
	cout << "Running Metropolis with " << max_steps << (time ? " seconds."  :  " steps.") << endl;
	
	// Main Metropolis step
	
	unsigned int mh_step(0);
	
	double used_time(0.0);
	
	double last_progress_dot_time = double (max_steps)/100.0;
	
	if (time)
	{
		global_timer.restart();
	}
	
	while (time ? used_time < max_steps : mh_step < max_steps )
	{
		++mh_step;
		
		if (!time)
		{
			if ( (mh_step % (max_steps/100) == 0) || (max_steps < 100) )
			{ 
				cout << "." << flush; 
			}
			
			if ( mh_step > burnoff_steps )
			{ 
				record = true; 
			}
		}
		
		else
		{
			if (used_time > last_progress_dot_time)
			{ 
				cout << "." << flush;
				last_progress_dot_time += double (max_steps)/100.0;
			}
			
			if ( used_time > burnoff_steps )
			{ 
				record = true; 
			}
			
			current_experiment->callback_monitoring(used_time);
		}
		
		for (tie (u,u_end) = vertices(g); u != u_end; ++u)
		{
			if ( out_degree (*u,g) == 1)
			{
				// We are in an observation node 
				continue;
			}
			
			//cout << "RV index " << g[*u]->get_random_variable()->get_index();
			
			current_random_variable = static_cast < ContinuousRandomVariable * > (g[*u]->get_random_variable());
			current_variable_index = current_random_variable->get_index();
			
			list <GaussianPotential *> edge_gaussian_lst;
			
			for (tie (v,v_end) = adjacent_vertices(*u,g); v != v_end; ++v)
			{
				if ( out_degree (*v,g) == 1)
				{
					// We are in an observation node, get the observation potential
					
					observation_potential = g[edge(*u,*v,g).first];
				}
				
				else 
				{
					edge_gaussian_lst.push_back( static_cast<GaussianPotential *> (g[edge(*u,*v,g).first]) );
				}
			}
			
			proposal.setup_gaussian_proposal (edge_gaussian_lst, current_variable_index);
			
			new_sample = proposal.get_proposal_position(*current_random_variable);
			
			pair <double, double> & previous_sample = get (previous_samples, *u);
			
			observation_potential->set_variable_value(*current_random_variable);
			
			new_observation_weight = observation_potential->get_potential_value();
			
			/*
			 if (new_observation_weight == 0.0)
			 {
				 cout << "Here too we reach 0 in OW" << endl;
			 }
			 */
			
			// previous_observation_weight = previous_sample.second;
			
			//previous_observation_weight = observation_potential->get_potential_value();
			
			if (*u == 19)
			{
				if (previous_sample.second == 0.0)
				{
					//cout << "arg " << mh_step << endl;
					
				}	
				else 
				{
					//cout << previous_sample.second<< " " << mh_step << endl;
				}
				
				
			}	
			
			double ratio = new_observation_weight / previous_sample.second;
			
			double z = uniform_distribution_0_1();
			
			if (z < ratio)
			{
				// We accept the new sample
				
				// Record it only if we are beyond the burnoff period
				
				if (record)
				{
					current_random_variable->record_sample(new_sample);
				}
				
				previous_sample.first = new_sample;
				previous_sample.second = new_observation_weight;
				
				// We need to update the potentials
				
				for (tie (v,v_end) = adjacent_vertices(*u,g); v != v_end; ++v)
				{
					if ( out_degree (*v,g) == 1)
					{
						// We are in an observation node 
						continue;
					}
					
					g[edge(*u,*v,g).first]->set_variable_value(*current_random_variable);
					
				}
				
			}
			
			else 
			{
				// Sample is rejected, we use the old one
				if (record)
				{
					current_random_variable->record_sample( previous_sample.first);
				}
				
			}
			
			
		}
		
		if (time)
		{
			used_time = global_timer.elapsed();
		}
		
	}
	
	for (tie (u,u_end) = vertices(g); u != u_end; ++u)
	{
		if ( out_degree (*u,g) == 1)
		{
			// We are in an observation node 
			continue;
		}
		
		g[*u]->get_random_variable()->obtain_inference_from_prior_samples();
	}
	
	cout << endl;
	
	if (time)
	{
		cout << "We used " << mh_step << " Metropolis-Hastings steps." << endl;
	}
}

