#include <ProbabilisticModel.h>

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/subgraph.hpp>
#include <boost/graph/copy.hpp>

#include <Node.h>
#include <RandomVariable.h>
#include <Potential.h>
#include <GraphAlgorithms.h>
#include <GraphTreePartition.h>
#include <OutputGraph.h>
#include <InputFunctions.h>
#include <Generic.h>

using namespace std;
using namespace boost;


bool get_next_begin_token(ifstream & input_file)
{
	char c = input_file.get();
	
	while ( !((c == '(') || (c == '<')) && (!input_file.eof()) )
	{
		c = input_file.get();
	}
	
	if (input_file.eof())
	{
		return false;
	}
	
	if (c == '<')
	{
		input_file.unget();
		return false;
	}
	
	return true;	
}



//***** ProbabilisticModel implementation ******//

ProbabilisticModel::~ProbabilisticModel()
{
}

//***** End of ProbabilisticModel implementation ******//

//***** DerivedProbabilisticModel implementation ******//

// Constructor

template <class G>
DerivedProbabilisticModel <G>::DerivedProbabilisticModel(const unsigned int n): g(n)
{
}

template <class G>
bool DerivedProbabilisticModel <G>::unserialize(const string & input_file_name)
{
	typedef typename graph_traits<G>::vertex_descriptor VertexDescriptor;
	
	string code_string;
	bool command_tag(false);
	
	ifstream input_file (input_file_name.c_str());
	
	if (!input_file.good())
	{
		cout << "Failed to open the file '" << input_file_name << "'" << endl;
		return false;
	}
	
	while (true)
	{		
		while ( (input_file.get() != '<') && (input_file.good()) )
		{
		}
		
		if (input_file.eof())
		{
			if (!command_tag)
			{
				cout << "No valid Command Tag found" << endl;
				return false;
			}
			
			break;
		}
		
		getline(input_file, code_string, '>');
		
		if ( code_string.compare("structure") == 0)
		{
			cout << "Found command tag 'structure'" << endl;
			command_tag = true;
			unserialize_structure(input_file);
		}		
		
		if ( code_string.compare("link") == 0)
		{
			cout << "Found command tag 'link'" << endl;
			command_tag = true;
			unserialize_links(input_file);
		}
		
		if ( code_string.compare("all random variable sizes") == 0)
		{
			cout << "Found command tag 'all random variable sizes'" << endl;
			command_tag = true;
			unserialize_all_random_variable_sizes(input_file);
		}
		
		if ( code_string.compare("random variable sizes") == 0)
		{
			cout << "Found command tag 'random variable sizes'" << endl;
			command_tag = true;
			unserialize_random_variable_sizes(input_file);
		}
		
		if ( code_string.compare("observations") == 0)
		{
			cout << "Found command tag 'observations'" << endl;
			command_tag = true;
			unserialize_observations(input_file);
		}
		
		if ( code_string.compare("potential tables") == 0)
		{
			cout << "Found command tag 'potential tables'" << endl;
			command_tag = true;
			unserialize_potential_tables(input_file);
		}
		
		if ( code_string.compare("internal potentials") == 0)
		{
			cout << "Found command tag 'internal potentials'" << endl;
			command_tag = true;
			unserialize_internal_potentials(input_file);
		}
		
		if ( code_string.compare("all potential tables") == 0)
		{
			cout << "Found command tag 'all potential tables'" << endl;
			command_tag = true;
			unserialize_all_potential_tables(input_file);
		}
		
		if ( code_string.compare("all internal potentials") == 0)
		{
			cout << "Found command tag 'all internal potentials'" << endl;
			command_tag = true;
			unserialize_all_internal_potentials(input_file);
		}
		
		if ( code_string.compare("ground truth") == 0)
		{
			cout << "Found command tag 'ground truth'" << endl;
			command_tag = true;
			if (!unserialize_ground_truth(input_file))
			{
				cout << "Parsing error occured in command tag 'ground truth'." << endl;
			}
		}		
	}
	
	return true;
}

template <class Graph>
bool DerivedProbabilisticModel <Graph>::unserialize_internal_potentials(ifstream & input_file)
{	
	typedef typename graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int a;
	
	while (true)
	{		
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		if (!(input_file >> a))
		{
			input_file.clear();
			cout << "Error: Malformed 'internal potentials' data tag" << endl;
			return false;
		}
		
		else
		{
			VertexDescriptor u = vertex(a-1, g);
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			while ( (input_file.get() != '(') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'internal potentials' data tag" << endl;
				return false;
			}
			
			vector < double> potential_table;
			double c;
			
			while (input_file >> c)
			{
				potential_table.push_back(c);
			}
				
			static_cast < DiscretePotential *> (g[u]->get_potential())->setup_potential_values(potential_table);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'internal potentials' data tag" << endl;
			return false;
		}	
	}
	
	return true;
}

template <class Graph>
bool DerivedProbabilisticModel <Graph>::unserialize_ground_truth(ifstream & original_input_file)
{	
	typedef typename graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int a;
	string code_string;
	
	ifstream external_input_file;
	ifstream * input_file_ptr = & original_input_file;
	
	if (!get_next_begin_token(original_input_file))
	{
		cout << 'a' << endl;
		return false;
	}
	
	getline(original_input_file, code_string, ')');
	
	if ( code_string.compare("external") == 0)
	{
		if (!get_next_begin_token(original_input_file))
		{
			cout << 'b' << endl;
			return false;
		}
		
		getline(original_input_file, code_string, ')');
		
		external_input_file.open(code_string.c_str());
		
		if (!external_input_file.good())
		{
			cout << "Failed to open the external file '" << code_string << "'" << endl;
			return false;
		}
		
		//cout << "Opened the external file '" << code_string << "'." << endl;
		
		while ( (external_input_file.get() != '<') && (external_input_file.good()) )
		{
		}
		
		if (external_input_file.eof())
		{
			cout << "Not a ground truth file." << endl;
			return false;
		}
		
		getline(external_input_file, code_string, '>');
		
		if (! code_string.compare("marginals") == 0)
		{
			cout << 'c' << endl;
			return false;
		}
		
		input_file_ptr = & external_input_file;
	}
	
	else
	{
		
		if (! code_string.compare("normal") == 0)
		{
			cout << 'd' << endl;
			return false;
		}
		
	}
	
	ifstream & input_file = * input_file_ptr;
	
	while (true)
	{
		if (!get_next_begin_token(input_file))
		{
			return true;
		}
		
		if (!(input_file >> a))
		{
			input_file.clear();
			cout << "Error: Malformed 'ground truth' data tag" << endl;
			return false;
		}
		
		else
		{
			VertexDescriptor u = vertex(a-1, g);
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			while ( (input_file.get() != '(') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'ground truth' data tag" << endl;
				return false;
			}
			
			vector < double> true_marginals;
			double c;
			
			while (input_file >> c)
			{
				true_marginals.push_back(c);
			}
			
			g[u]->get_random_variable()->set_reference_marginals(true_marginals);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'ground truth' data tag" << endl;
			return false;
		}		
		
	}
	
	return true;
}



template <class G>
void DerivedProbabilisticModel <G>::get_inference_mp()
{	
	belief_propagation(g);
}

template <class G>
void DerivedProbabilisticModel <G>::get_inference_loopy(const unsigned int a, const bool timer)
{
	loopy_belief_propagation(g, a, timer);
}

template <class G>
void DerivedProbabilisticModel <G>::get_inference_gibbs(const unsigned int gibbs_steps, const bool timer )
{
	gibbs_sampler(g, gibbs_steps, 0, timer);
}

template <class G>
void DerivedProbabilisticModel <G>::get_inference_tree_gibbs(const unsigned int gibbs_steps, const bool rao_blackwell, const bool timer)
{
	tree_gibbs_sampler(g,  gibbs_steps, 0, rao_blackwell, timer);
}

template <class G>
void DerivedProbabilisticModel <G>::get_tree_partition()
{
	subgraph <G> subgraph_g;
	copy_graph(g, subgraph_g);
	partition_graph_into_trees(subgraph_g);
	display_subgraph(subgraph_g);
	
	unsigned int a(0);
	
	typename subgraph<G>::children_iterator s, s_end;	
	
	for ( tie(s, s_end) = subgraph_g.children(); s != s_end; ++s)
	{
		++a;
	}
	
	number_of_trees.push_back(a);
}

template <class G>
bool DerivedProbabilisticModel <G>::serialize_marginals(ofstream & output_file)
{
	output_file << "<marginals> " ;
	
	typename graph_traits<G>::vertex_iterator u, u_end;
	
	for (tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		g[*u]->serialize_marginals(output_file);
	}
	
	output_file << endl;
	
	return true;
}

template <class G>
bool DerivedProbabilisticModel <G>::serialize_partition(ofstream & output_file)
{
	output_file << "<partition> ";
	
	typename graph_traits<G>::vertex_iterator u, u_end;
	
	for (vector<unsigned int>::iterator it = number_of_trees.begin(); it != number_of_trees.end(); ++it)
	{
		output_file << *it << " ";
	}
	
	output_file << endl;
	
	return true;
}

template <class G>
bool DerivedProbabilisticModel <G>::serialize(ofstream & output_file)
{
	return serialize_marginals(output_file);
}

//***** End of DerivedProbabilisticModel implementation ******//

//***** DiscreteProbabilisticModel implementation ******//

// Constructor

DiscreteProbabilisticModel::DiscreteProbabilisticModel(unsigned int n): DerivedProbabilisticModel<DiscretePairwiseGraph>(n)
{
}

// Virtual functions

bool DiscreteProbabilisticModel::unserialize_structure(ifstream & input_file)
{	
	typedef graph_traits<DiscretePairwiseGraph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	double c;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("square-lattice") == 0)
	{
		while ( (input_file.get() != '(')  && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error, no parameters given to square-lattice" << endl;
			return false;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'structure' data tag" << endl;
			return false;
		}
		else
		{
			create_pairwise_square_lattice(a, b, g);
		}
		
	}
	
	
	if (code_string.compare("random") == 0)
	{
		while ( (input_file.get() != '(')  && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error, no parameters given to random" << endl;
			return false;
		}
		
		if (!(input_file >> c))
		{
			input_file.clear();
			cout << "Error: Malformed 'structure' data tag" << endl;
			return false;
		}
		else
		{
			create_random_graph(c, g);
		}
	}	
	
	input_file.clear();
	
	while ( (input_file.get() != ')') && (!input_file.eof()) )
	{
	}
	
	if (input_file.eof())
	{
		cout << "Error: malformed 'link' data tag" << endl;
		return false;
	}
	
	
	return true;
}

bool DiscreteProbabilisticModel::unserialize_all_random_variable_sizes(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	while (true)
	{
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		if (!(input_file >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'link' data tag" << endl;
			return false;
		}
		else
		{
			a = 0;
			graph_traits<Graph>::vertex_iterator u, u_end;
			
			for (tie (u, u_end) = vertices(g); u != u_end; ++u)
			{
				++a;
				
				DiscreteRandomVariable * rv;
				MRFMessageNode <DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> * node;
				
				rv = new DiscreteRandomVariable(a, b);
				node = new MRFMessageNode<DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor>(rv, new SingleDiscretePotential (*rv));
				
				g[*u] = node;
			}
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'link' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}


bool DiscreteProbabilisticModel::unserialize_random_variable_sizes(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	while (true)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'link' data tag" << endl;
			return false;
		}
		else
		{
			
			VertexDescriptor u = vertex(a-1, g);
			
			DiscreteRandomVariable * rv;
			MRFMessageNode <DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> * node;
			
			rv = new DiscreteRandomVariable(a, b);
			node = new MRFMessageNode<DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor>(rv, new SingleDiscretePotential (*rv));
			
			g[u] = node;
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'link' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}

bool DiscreteProbabilisticModel::unserialize_observations(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	while (true)
	{		
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'observations' data tag" << endl;
			return false;
		}
		
		else
		{
			
			VertexDescriptor u = vertex(a-1, g);
			g[u]->get_random_variable()->observe_variable_value(b);
			
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'observations' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}


bool DiscreteProbabilisticModel::unserialize_links(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int a, b;
	
	while (true)
	{
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'link' data tag" << endl;
			return false;
		}
		
		else
		{
			VertexDescriptor u = vertex(a-1, g);
			VertexDescriptor v = vertex(b-1, g);
			
			graph_traits<Graph>::edge_descriptor e = add_edge(u,v,g).first;	
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'link' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}



bool DiscreteProbabilisticModel::unserialize_all_potential_tables(ifstream & input_file)
{		
	graph_traits<Graph>::edge_iterator e, e_end;
	
	vector < double> potential_table;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("all-random") == 0)
	{
		for ( tie (e,e_end) = edges(g); e != e_end; ++e)
		{
			
			unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
			unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
			
			potential_table.resize(a*b);
			
			for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
			{
				*it= uniform_distribution_0_1();	
			}
			
			g[*e] = new DiscretePotential();
			g[*e]->add_variable(* g[source(*e, g)]->get_random_variable());
			g[*e]->add_variable(* g[target(*e, g)]->get_random_variable());
			
			g[*e]->setup_potential_values(potential_table);
		}
		
		return true;
	}
	
	
	if (code_string.compare("all-random-strongly-diagonal-uniform-noise") == 0)
	{
		
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'all-random-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}
		
		for ( tie (e,e_end) = edges(g); e != e_end; ++e)
		{
			
			unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
			unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
			
			if (a != b)
			{
				cout << "The potential here is not a squared  matrix, I cannot make it strongly diagonal with noise." << endl;
				return false;
			}
			
			potential_table.resize(a*b);
			
			for (unsigned int i = 0; i < potential_table.size(); ++i)
			{
				if ( (i % a) == ( i/a ) )
				{
					potential_table[i] = uniform_distribution_0_1()*strongly_diagonal_coefficient;
				}
				
				else
				{
					potential_table[i] = 1.0;
				}
				
			}
			
			g[*e] = new DiscretePotential();
			g[*e]->add_variable(* g[source(*e, g)]->get_random_variable());
			g[*e]->add_variable(* g[target(*e, g)]->get_random_variable());
			
			g[*e]->setup_potential_values(potential_table);
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-random-noise' data tag in 'all potential tables' tag" << endl;
			return false;
		}	
		
		return true;
	}
	
	if (code_string.compare("normal") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double a;
		
		while (input_file >> a)
		{
			potential_table.push_back(a);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'normal' data tag in 'all potential tables' tag" << endl;
			return false;
		}
		
	}
	
	// Following will create one random potential and use it for all the potentials in the graph
	
	if (code_string.compare("one-random") == 0)
	{
		tie (e,e_end) = edges(g);
		
		unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
		unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
		
		potential_table.resize(a*b);
		
		for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
		{
			*it= uniform_distribution_0_1();	
		}		
	}
	
	// Here we create a one strongly diagonal potential and use it for all the potentials in the graph
	
	if (code_string.compare("one-random-strongly-diagonal-random-noise") == 0)
	{
		tie (e,e_end) = edges(g);
		
		unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
		unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
		
		if (a != b)
		{
			cout << "The potential here is not a squared  matrix, I cannot make it strongly diagonal with noise." << endl;
			return false;
		}
		
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-random-noise' data tag in 'all potential tables' tag" << endl;
			return false;
		}
		
		potential_table.resize(a*b);
		
		for (unsigned int i = 0; i < potential_table.size(); ++i)
		{
			if ( (i % a) == ( i/a ) )
			{
				potential_table[i] = uniform_distribution_0_1()*strongly_diagonal_coefficient;
			}
			
			else
			{
				potential_table[i] = uniform_distribution_0_1();
			}
			
		}
		
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-random-noise' data tag in 'all potential tables' tag" << endl;
			return false;
		}	
		
	}
	
	if (code_string.compare("one-random-strongly-diagonal-uniform-noise") == 0)
	{
		tie (e,e_end) = edges(g);
		
		unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
		unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
		
		if (a != b)
		{
			cout << "The potential here is not a squared  matrix, I cannot make it strongly diagonal with noise." << endl;
			return false;
		}
		
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-uniform-noise' data tag in 'all potential tables' tag" << endl;
			return false;
		}
		
		potential_table.resize(a*b);
		
		for (unsigned int i = 0; i < potential_table.size(); ++i)
		{
			if ( (i % a) == ( i/a ) )
			{
				potential_table[i] = uniform_distribution_0_1()*strongly_diagonal_coefficient;
			}
			
			else
			{
				potential_table[i] = 1.0;
			}
			
		}
		
		/*
		 for (unsigned int i = 0; i < potential_table.size(); ++i)
		 {
			 cout << potential_table[i] << " ";
		 }
		 */
		
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-uniform-noise' data tag in 'all potential tables' tag" << endl;
			return false;
		}	
		
	}
	
	if (code_string.compare("one-uniform-strongly-diagonal-uniform-noise") == 0)
	{
		tie (e,e_end) = edges(g);
		
		unsigned int a = g[source(*e, g)]->get_random_variable()->get_number_values();
		unsigned int b = g[target(*e, g)]->get_random_variable()->get_number_values();
		
		if (a != b)
		{
			cout << "The potential here is not a squared  matrix, I cannot make it strongly diagonal with noise." << endl;
			return false;
		}
		
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'one-uniform-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}
		
		potential_table.resize(a*b);
		
		for (unsigned int i = 0; i < potential_table.size(); ++i)
		{
			if ( (i % a) == ( i/a ) )
			{
				potential_table[i] = strongly_diagonal_coefficient;
			}
			
			else
			{
				potential_table[i] = 1.0;
			}
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'one-uniform-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}	
		
	}
	
	for ( tie (e,e_end) = edges(g); e != e_end; ++e)
	{
		g[*e] = new UniqueDiscretePotential();
		g[*e]->add_variable(* g[source(*e, g)]->get_random_variable());
		g[*e]->add_variable(* g[target(*e, g)]->get_random_variable());
		
		//cout << "Edge between " << source(*e, g) << " and " << target(*e, g) << endl;
		
		g[*e]->setup_potential_values(potential_table);
	}
	
	//cout << "Done with potential tables" << endl;
	return true;
}

bool DiscreteProbabilisticModel::unserialize_all_internal_potentials(ifstream & input_file)
{		
	graph_traits<Graph>::vertex_iterator u, u_end;
	
	vector < double> potential_table;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("random-correlated") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double coefficient(1.0);
		
		if (!(input_file >> coefficient))
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}
		
		for ( tie (u,u_end) = vertices(g); u != u_end; ++u)
		{
			unsigned int a = g[*u]->get_random_variable()->get_number_values();
			
			potential_table.resize(a);
			fill(potential_table.begin(), potential_table.end(), 1.0);
			
			uniform_int <int> uni_int_dist(0, a-1);
			
			variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg (fibonnacci_number_generator, uni_int_dist);
			
			potential_table[vg()] = coefficient;
			
			static_cast <DiscretePotential *> (g[*u]->get_potential())->setup_potential_values(potential_table);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}	
		
	}
	
	if (code_string.compare("special") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double coefficient(1.0);
		
		if (!(input_file >> coefficient))
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}
		
		for ( tie (u,u_end) = vertices(g); u != u_end; ++u)
		{
			unsigned int a = g[*u]->get_random_variable()->get_number_values();
			
			potential_table.resize(a);
			fill(potential_table.begin(), potential_table.end(), 1.0);
			
			uniform_int <int> uni_int_dist(0, 1);
			
			variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg (fibonnacci_number_generator, uni_int_dist);
			
			potential_table[vg()] = coefficient;
			
			static_cast <DiscretePotential *> (g[*u]->get_potential())->setup_potential_values(potential_table);
			
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}	
		
	}
	
	return true;
}


bool DiscreteProbabilisticModel::unserialize_potential_tables(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int a, b;
	
	while (true)
	{		
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'potential tables' data tag" << endl;
			return false;
		}
		
		else
		{
			VertexDescriptor u = vertex(a-1, g);
			VertexDescriptor v = vertex(b-1, g);
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			while ( (input_file.get() != '(') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'potential tables' data tag" << endl;
				return false;
			}
			
			vector < double> potential_table;
			double c;
			
			while (input_file >> c)
			{
				potential_table.push_back(c);
			}
			
			graph_traits<Graph>::edge_descriptor e = edge(u,v,g).first;
			
			g[e] = new DiscretePotential();
			g[e]->add_variable(* g[u]->get_random_variable());
			g[e]->add_variable(* g[v]->get_random_variable());
			
			g[e]->setup_potential_values(potential_table);
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'potential tables' data tag" << endl;
			return false;
		}
		
		
	}
	
	return true;
}

double DiscreteProbabilisticModel::compute_error_to_reference()
{		
	graph_traits<Graph>::vertex_iterator u, u_end;
	
	double error(0.0);
	
	for ( tie(u, u_end) = vertices(g); u != u_end; ++u)
	{
		error += g[*u]->get_random_variable()->compute_error_to_reference();
	}
	
	return error;
}

//***** End of DiscreteProbabilisticModel implementation ******//

//***** DiscreteFactorGraphProbabilisticModel implementation ******//

// Constructor

DiscreteFactorGraphProbabilisticModel::DiscreteFactorGraphProbabilisticModel(unsigned int n): DerivedProbabilisticModel<DiscreteFactorGraph>(n), number_random_variables(0)
{
}

// Inherited methods

bool DiscreteFactorGraphProbabilisticModel::unserialize_all_random_variable_sizes(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int size;
	
	if(!get_next_begin_token(input_file))
	{
		return false;
	}
	
	if (!(input_file >> number_random_variables >> size))
	{
		input_file.clear();
		cout << "Error: Malformed 'all random variable sizes' data tag" << endl;
		return false;
	}
	
	else
	{
		
		for (unsigned int i = 1; i <= number_random_variables; ++i)
		{
			//cout << i << endl;
			VertexDescriptor u = vertex(i-1, g);
			
			DiscreteRandomVariable * rv;
			VariableMessageNode <DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> * node;
			
			rv = new DiscreteRandomVariable(i, size);
			node = new VariableMessageNode<DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor>(rv, new SingleDiscretePotential (*rv));
			
			//cout << "Creating a Variable Node with internal index " << u+1 << endl;
			
			g[u] = node;
		}
		
	}
	
	while ( (input_file.get() != ')') && (!input_file.eof()) )
	{
	}
	
	if (input_file.eof())
	{
		cout << "Error: malformed 'all random variable sizes' data tag" << endl;
		return false;
	}
	
	//cout << "ok" << endl;
	
	return true;
}


bool DiscreteFactorGraphProbabilisticModel::unserialize_all_potential_tables(ifstream & input_file)
{
	graph_traits<Graph>::adjacency_iterator v, v_end;
	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	vector < double> potential_table;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("all-random") == 0)
	{
		for ( unsigned int i = number_random_variables; i < num_vertices(g) ; ++i)
		{
			VertexDescriptor u = vertex(i, g);
			
			unsigned int a = 1;
			
			for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
			{
				a = a * g[*v]->get_random_variable()->get_number_values();
			}
			
			potential_table.resize(a);
			
			for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
			{
				*it= uniform_distribution_0_1();	
			}
			
			static_cast < DiscretePotential * > (g[u]->get_potential())->setup_potential_values(potential_table);
		}
		
		return true;
	}
	
	// Here we create a one strongly diagonal potential and use it for all the potentials in the graph
	
	
	if (code_string.compare("all-random-strongly-diagonal-random-noise") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'one-random-strongly-diagonal-random-noise' data tag" << endl;
			return false;
		}
		
		for ( unsigned int i = number_random_variables; i < num_vertices(g) ; ++i)
		{
			
			VertexDescriptor u = vertex(i, g);
			
			unsigned int a = 1;
			
			for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
			{
				a = a * g[*v]->get_random_variable()->get_number_values();
			}
			
			potential_table.resize(a);
			
			for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
			{
				*it= uniform_distribution_0_1();	
			}
			
			static_cast < DiscretePotential * > (g[u]->get_potential())->setup_potential_values(potential_table);
			
			tie (v,v_end) = adjacent_vertices(u,g);
			
			VertexDescriptor w = *v;
			
			while (g[w]->get_random_variable()->loop_over())
			{
				for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
				{
					g[*v]->get_random_variable()->set_equal_to_variable( * g[w]->get_random_variable());
					g[u]->get_potential()->set_variable_value ( * g[*v]->get_random_variable());
				}
				
				static_cast < DiscretePotential * > (g[u]->get_potential())->modify_potential_value( uniform_distribution_0_1()*strongly_diagonal_coefficient);
			}
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'all-random-strongly-diagonal-random-noise' data tag" << endl;
			return false;
		}	
		
		return true;
	}
	
	if (code_string.compare("all-random-strongly-diagonal-uniform-noise") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'all-random-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}
		
		for ( unsigned int i = number_random_variables; i < num_vertices(g) ; ++i)
		{
		//cout << "Potential " << i+1 << endl;
			
			VertexDescriptor u = vertex(i, g);

			unsigned int a = 1;
			
			for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
			{
				a = a * g[*v]->get_random_variable()->get_number_values();
			}
			
			potential_table.resize(a);
			
			for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
			{
				*it= 1.0;	
			}
			
			static_cast < DiscretePotential * > (g[u]->get_potential())->setup_potential_values(potential_table);
			
			tie (v,v_end) = adjacent_vertices(u,g);
			
			VertexDescriptor w = *v;
			
			while (g[w]->get_random_variable()->loop_over())
			{
				for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
				{
					g[*v]->get_random_variable()->set_equal_to_variable( * g[w]->get_random_variable());
					g[u]->get_potential()->set_variable_value ( * g[*v]->get_random_variable());
				}
				
				
				static_cast < DiscretePotential * > (g[u]->get_potential())->modify_potential_value( uniform_distribution_0_1()*strongly_diagonal_coefficient);
			}
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'all-random-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}	
		
		return true;
	}
	
	if (code_string.compare("all-uniform-strongly-diagonal-uniform-noise") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double strongly_diagonal_coefficient(1.0);
		
		if (!(input_file >> strongly_diagonal_coefficient))
		{
			cout << "Error: malformed 'one-uniform-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}
		
		for ( unsigned int i = number_random_variables; i < num_vertices(g) ; ++i)
		{
			VertexDescriptor u = vertex(i, g);
			
			unsigned int a = 1;
			
			for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
			{
				a = a * g[*v]->get_random_variable()->get_number_values();
			}
			
			potential_table.resize(a);
			
			for (vector<double>::iterator it = potential_table.begin(); it !=  potential_table.end(); ++it)
			{
				*it= 1.0;	
			}
			
			static_cast < DiscretePotential * > (g[u]->get_potential())->setup_potential_values(potential_table);
			
			tie (v,v_end) = adjacent_vertices(u,g);
			
			VertexDescriptor w = *v;
			
			while (g[w]->get_random_variable()->loop_over())
			{
				for ( tie (v,v_end) = adjacent_vertices(u,g); v!= v_end; ++v)
				{
					g[*v]->get_random_variable()->set_equal_to_variable( * g[w]->get_random_variable());
					g[u]->get_potential()->set_variable_value ( * g[*v]->get_random_variable());
				}
				
				
				static_cast < DiscretePotential * > (g[u]->get_potential())->modify_potential_value( strongly_diagonal_coefficient);
			}
			
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'all-uniform-strongly-diagonal-uniform-noise' data tag" << endl;
			return false;
		}	
		
		return true;
	}
	
	return true;	
}

bool DiscreteFactorGraphProbabilisticModel::unserialize_structure(ifstream & input_file)
{
	unsigned int a, b;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("random") == 0)
	{
		while ( (input_file.get() != '(')  && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error, no parameters given to random" << endl;
			return false;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'structure' data tag" << endl;
			return false;
		}
		else
		{
			number_random_variables = a;
			
			create_random_graph(b);
		}
	}	
	
	input_file.clear();
	
	while ( (input_file.get() != ')') && (!input_file.eof()) )
	{
	}
	
	if (input_file.eof())
	{
		cout << "Error: malformed 'link' data tag" << endl;
		return false;
	}
	
	
	return true;
	
}

bool DiscreteFactorGraphProbabilisticModel::unserialize_random_variable_sizes(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	while (true)
	{		
		char c = input_file.get();
		
		while ( !((c == '(') || (c == '<')) && (!input_file.eof()) )
		{
			c = input_file.get();
		}
		
		if (input_file.eof())
		{
			break;
		}
		
		if (c == '<')
		{
			input_file.unget();
			break;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'RV size' data tag" << endl;
			return false;
		}
		else
		{
			++number_random_variables;
			
			VertexDescriptor u = vertex(a-1, g);
			
			DiscreteRandomVariable * rv;
			VariableMessageNode <DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> * node;
			
			rv = new DiscreteRandomVariable(a, b);
			node = new VariableMessageNode<DiscreteRandomVariable, DiscretePotential, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor>(rv, new SingleDiscretePotential (*rv));
			
			cout << "Creating a Variable Node with internal index " << u+1 << endl;
			
			g[u] = node;
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'link' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}

bool DiscreteFactorGraphProbabilisticModel::unserialize_observations(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int a, b;
	
	while (true)
	{
		char c = input_file.get();
		
		while ( !((c == '(') || (c == '<')) && (!input_file.eof()) )
		{
			c = input_file.get();
		}
		
		if (input_file.eof())
		{
			break;
		}
		
		if (c == '<')
		{
			input_file.unget();
			break;
		}
		
		if (!(input_file >> a >> b))
		{
			input_file.clear();
			cout << "Error: Malformed 'observations' data tag" << endl;
			return false;
		}
		else
		{
			
			VertexDescriptor u = vertex(a-1, g);
			g[u]->get_random_variable()->observe_variable_value(b);
			
		}
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'observations' data tag" << endl;
			return false;
		}
		
	}
	
	return true;
}


bool DiscreteFactorGraphProbabilisticModel::unserialize_links(ifstream & input_file)
{	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int current_potential_node(number_random_variables);
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("normal") == 0)
	{

	while (true)
	{		
		if (!get_next_begin_token(input_file))
		{
			break;
		}
		
		unsigned int a(0);
		vector <unsigned int> link_vec;
		
		while (input_file >> a)
		{
			link_vec.push_back(a);
		}
		
		input_file.clear();
		
		++current_potential_node;
		
		VertexDescriptor u = vertex(current_potential_node-1, g); // current potential node
		
		DiscretePotential * pot_ptr = new DiscretePotential();
		
		for (vector <unsigned int>::iterator it = link_vec.begin(); it != link_vec.end(); ++it)
		{
			VertexDescriptor v = vertex(*it-1, g);
			pot_ptr->add_variable(* g[v]->get_random_variable());
			
			//cout << "Adding edge: " << u << " and " << v << endl;
			
			add_edge(u,v,g);
		}
		
		// It is probably wrong to create the MN now when the Potential is not yet fully built
		
		cout << "Creating a Potential Node with internal index " << u+1 << endl;
		
		g[u] = new PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, VertexDescriptor> (current_potential_node, pot_ptr);
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'link' data tag" << endl;
			return false;
		}
		
	}
		
	}
	
	if (code_string.compare("qmr") == 0) 
	{
		while (true)
		{		
			if (!get_next_begin_token(input_file))
			{
				break;
			}
			
			unsigned int a(0);
			vector <unsigned int> link_vec;
			
			while (input_file >> a)
			{
				link_vec.push_back(a);
			}
			
			input_file.clear();
			
			++current_potential_node;
			
			VertexDescriptor u = vertex(current_potential_node-1, g); // current potential node
			
			DiscretePotential * pot_ptr = new QMRDiscretePotential();
			
			for (vector <unsigned int>::iterator it = link_vec.begin(); it != link_vec.end(); ++it)
			{
				VertexDescriptor v = vertex(*it-1, g);
				pot_ptr->add_variable(* g[v]->get_random_variable());
				
				//cout << "Adding edge: " << u << " and " << v << endl;
				
				add_edge(u,v,g);
			}
			
			// It is probably wrong to create the MN now when the Potential is not yet fully built
			
			cout << "Creating a QMR Potential Node with internal index " << u+1 << endl;
			
			g[u] = new PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, VertexDescriptor> (current_potential_node, pot_ptr);
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'link' data tag" << endl;
				return false;
			}
			
		}
		
	}
	
	return true;
}

bool DiscreteFactorGraphProbabilisticModel::unserialize_potential_tables(ifstream & input_file)
{
	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int current_potential = number_random_variables;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if ( (code_string.compare("normal") == 0) || (code_string.compare("qmr") == 0) )
	{
		
		while (true)
		{		
			if (!get_next_begin_token(input_file))
			{
				break;
			}
			
			++current_potential;
			
			VertexDescriptor u = vertex(current_potential-1, g);
			
			vector < double> potential_table;
			double a;
			
			while (input_file >> a)
			{
				potential_table.push_back(a);
			}
			
			static_cast < DiscretePotential *> (g[u]->get_potential() )->setup_potential_values(potential_table);
			
			input_file.clear();
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'potential tables' data tag" << endl;
				return false;
			}
			
		}
		
		
		
	}
	
	return true;	
}

/*
bool DiscreteFactorGraphProbabilisticModel::unserialize_internal_potentials(std::ifstream &)
{
	
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	unsigned int current_rv = 0;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if ( (code_string.compare("normal") == 0) || (code_string.compare("qmr") == 0) )
	{
		
		while (true)
		{		
			if (!get_next_begin_token(input_file))
			{
				break;
			}
			
			
			VertexDescriptor u = vertex(current_rv, g);
			
			vector < double> potential_table;
			double a;
			
			while (input_file >> a)
			{
				potential_table.push_back(a);
			}
			
			static_cast < DiscretePotential *> (g[u]->get_potential() )->setup_potential_values(potential_table);
			
			input_file.clear();
			
			while ( (input_file.get() != ')') && (!input_file.eof()) )
			{
			}
			
			if (input_file.eof())
			{
				cout << "Error: malformed 'potential tables' data tag" << endl;
				return false;
			}
			
			
			++current_rv;
			
		}
		
		
		
	}
	
	return true;	
	
	
}
*/

bool DiscreteFactorGraphProbabilisticModel::unserialize_all_internal_potentials(ifstream & input_file)
{
	
	vector < double> potential_table;
	
	string code_string;
	
	if (!get_next_begin_token(input_file))
	{
		return false;
	}
	
	getline(input_file, code_string, ')');
	
	if (code_string.compare("random-correlated") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double coefficient(1.0);
		
		if (!(input_file >> coefficient))
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}
		
		for ( unsigned int i = 0; i < number_random_variables; ++i)
		{
			graph_traits<DiscreteFactorGraph>::vertex_descriptor u = vertex(i,g);
			
			unsigned int a = g[u]->get_random_variable()->get_number_values();
			
			potential_table.resize(a);
			fill(potential_table.begin(), potential_table.end(), 1.0);
			
			uniform_int <int> uni_int_dist(0, a-1);
			
			variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg (fibonnacci_number_generator, uni_int_dist);
			
			potential_table[vg()] = coefficient;
			
			static_cast <DiscretePotential *> (g[u]->get_potential())->setup_potential_values(potential_table);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'random-correlated' data tag" << endl;
			return false;
		}	
		
	}
	
	if (code_string.compare("special") == 0)
	{
		if (!get_next_begin_token(input_file))
		{
			return false;
		}
		
		double coefficient(1.0);
		
		if (!(input_file >> coefficient))
		{
			cout << "Error: malformed 'special' data tag" << endl;
			return false;
		}
		
		for ( unsigned int i = 0; i < number_random_variables; ++i)
		{
			graph_traits<DiscreteFactorGraph>::vertex_descriptor u = vertex(i,g);
			
			unsigned int a = g[u]->get_random_variable()->get_number_values();
			
			potential_table.resize(a);
			fill(potential_table.begin(), potential_table.end(), 1.0);
			
			uniform_int <int> uni_int_dist(0, 1);
			
			variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg (fibonnacci_number_generator, uni_int_dist);
			
			potential_table[vg()] = coefficient;
			
			static_cast <DiscretePotential *> (g[u]->get_potential())->setup_potential_values(potential_table);
		}
		
		input_file.clear();
		
		while ( (input_file.get() != ')') && (!input_file.eof()) )
		{
		}
		
		if (input_file.eof())
		{
			cout << "Error: malformed 'special' data tag" << endl;
			return false;
		}	
		
	}
	
	
	return true;
}

double DiscreteFactorGraphProbabilisticModel::compute_error_to_reference()
{
	graph_traits<Graph>::vertex_iterator u, u_end;
	
	double error(0.0);
	
	for ( tie(u, u_end) = vertices(g); u != u_end; ++u)
	{
		VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<Graph>::vertex_descriptor> * vmn = dynamic_cast < VariableMessageNode <DiscreteRandomVariable, DiscretePotential, graph_traits<Graph>::vertex_descriptor> *> ( g[*u]);
		
		if (vmn != NULL)
		{
			error += vmn->get_random_variable()->compute_error_to_reference();
		}
	}
	
	return error;
}

void DiscreteFactorGraphProbabilisticModel::create_random_graph(const unsigned int density)
{
	typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
	
	unsigned int number_potentials = num_vertices(g) - number_random_variables;
	cout << 2 << "-" << density << endl;
	uniform_int <int> uni_int_dist(2, density);
	
	
	variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg (fibonnacci_number_generator, uni_int_dist);
	
	cout <<  0 << "-" << number_random_variables-1 << endl;
	
	uniform_int <int> uni_int_dist_2(0, number_random_variables-1);
	
	//vector <DiscretePotential *> pot_ptr; 
	
	variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg_2 (fibonnacci_number_generator, uni_int_dist_2);
	
	for (unsigned int i = 0; i < number_potentials; ++i)
	{
		VertexDescriptor u = vertex(i+number_random_variables, g); // current potential node
		
		//pot_ptr.push_back(new DiscretePotential());
		
		unsigned int a = vg();
		
		for (unsigned int j = 0; j < a;++j)
		{
			VertexDescriptor v = vertex(vg_2(), g);
			
			if (!edge(u,v,g).second)
			{
				add_edge(u,v,g);
				//pot_ptr->add_variable(* g[v]->get_random_variable());				
			}
			
			assert(edge(u,v,g).second);
		}
		
		//g[u] = new PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, VertexDescriptor> (number_random_variables+1+i, pot_ptr);
		
	}
	cout <<  number_random_variables << "-" << number_random_variables+number_potentials-1 << endl;
	uniform_int <int> uni_int_dist_3(number_random_variables, number_random_variables+number_potentials-1);
	
	variate_generator <boost::lagged_fibonacci607 &, uniform_int <int> > vg_3 (fibonnacci_number_generator, uni_int_dist_3);
	
	for (unsigned int i = 0; i < number_random_variables; ++i)
	{
		VertexDescriptor u = vertex(i, g); // current variable node
		
		if (out_degree(u,g) == 0)
		{
			VertexDescriptor v = vertex(vg_3(), g); // current variable node
			//static_cast <DiscretePotential * >(g[v]->get_potential())->add_variable(* g[u]->get_random_variable());
			add_edge(u,v,g);
		}
		
	}
	
	for (unsigned int i = 0; i < number_potentials; ++i)
	{
		VertexDescriptor u = vertex(i+number_random_variables, g); // current potential node
		
		DiscretePotential * pot_ptr = new DiscretePotential();
		
		graph_traits<Graph>::adjacency_iterator v, v_end;
		
		for (tie (v, v_end) = adjacent_vertices(u,g); v != v_end; ++v)
		{
			pot_ptr->add_variable(* g[*v]->get_random_variable());
		}
		
		g[u] = new PotentialMessageNode <DiscreteRandomVariable, DiscretePotential, VertexDescriptor> (number_random_variables+1+i, pot_ptr);
		
	}
	
	// Below ensures that we only have ONE connected component in our Random Graph.
	/*
	 std::vector<unsigned int> component(number_of_nodes);
	 unsigned int number_of_components = connected_components(g, &component[0]);
	 
	 std::vector<unsigned int>::size_type i;
	 
	 vector <unsigned int> component_contents (number_of_components);
	 
	 for (i = 0; i != number_of_nodes; ++i)
	 {
		 component_contents[component[i]] = i;
	 }
	 
	 for (i = 0; i < number_of_components -1 ; ++i)
	 {
		 t = vertex(component_contents[i],g);
		 u = vertex(component_contents[i+1],g);
		 add_edge (t, u, g);
	 }
	 
	 assert (connected_components(g, &component[0]) == 1);
	 */
	
	display_graph(g);
}

//***** End of DiscreteFactorGraphProbabilisticModel implementation ******//
